#!/usr/bin/python
import os, sys
from io import StringIO
from django.conf import settings
import imaplib
import email
from email.parser import Parser as EmailParser
from email.utils import parseaddr, parsedate
from email.header import decode_header
import time
from datetime import datetime
import tempfile
from rest.helpers import toString
from rest.uberdict import UberDict
from optparse import OptionParser

# lazy log
def log(msg):
    # print "{0}\t{1}".format(datetime.now(), msg)
    # sys.stdout.flush()
    pass

import base64
import xmltodict

class NotSupportedMailFormat(Exception):
    pass

def parse_attachment(message_part):
    content_disposition = message_part.get("Content-Disposition", None)
    if content_disposition:
        dispositions = content_disposition.strip().split(";")
        if bool(content_disposition and dispositions[0].lower() == "attachment"):
            file_data = message_part.get_payload(decode=True)
            # Used a StringIO object since PIL didn't seem to recognize
            # images using a custom file-like object
            attachment = StringIO(toString(file_data))
            attachment.content_type = message_part.get_content_type()
            attachment.size = len(file_data)
            attachment.name = None
            attachment.create_date = None
            attachment.mod_date = None
            attachment.read_date = None
            # print dispositions
            for param in dispositions[1:]:
                name,value = param.split("=")
                name = name.strip().lower()
                if name == "filename":
                    attachment.name = value
                elif name in ["create-date", "creation-date"]:
                    attachment.create_date = value  #TODO: datetime
                elif name == "modification-date":
                    attachment.mod_date = value #TODO: datetime
                elif name == "read-date":
                    attachment.read_date = value #TODO: datetime
            return attachment
    return None

def parse_raw_message(msgobj):
    """
    Parse the email and return a dictionary of relevant data.
    """
    if type(msgobj) in [str, str]:
        msgobj = email.message_from_string(msgobj)
    if msgobj['Subject'] is not None:
        decodefrag = decode_header(msgobj['Subject'])
        subj_fragments = []
        for s, enc in decodefrag:
            if enc:
                s = str(s, enc).encode('utf8', 'replace')
            subj_fragments.append(s)
        subject = ''.join(subj_fragments)
    else:
        subject = None
    attachments = []
    body = None
    html = None
    for part in msgobj.walk():
        attachment = parse_attachment(part)
        if attachment:
            attachments.append(attachment)
        elif part.get_content_type() == "text/plain":
            if body is None:
                body = ""
            try:
                body += str(
                    part.get_payload(decode=True),
                    part.get_content_charset(),
                    'replace'
                )
            except Exception:
                body += str(part.get_payload(decode=True))

        elif part.get_content_type() == "text/html":
            if html is None:
                html = ""
            html += str(
                part.get_payload(decode=True),
                part.get_content_charset(),
                'replace'
            )
    if html:
        html = html.strip()
    message = []
    if body:
        body = body.strip()
        # now lets parse the first part of the message that is not "quoted"
        blocks = 0
        for line in body.split('\n'):
            if line.startswith('>'):
                blocks += 1
                if blocks > 3:
                    break
                continue
            blocks = 0
            message.append(line.strip())
    message = "\n".join(message).strip()
    from_addr = parseaddr(msgobj.get('From'))
    to_addr = parseaddr(msgobj.get('To'))
    date_time = parsedate(msgobj.get('Date'))
    return UberDict({
        'subject' : subject.strip(),
        'body' : body,
        'date_time' : date_time,
        'message': message,
        'html' : html,
        'from_email': from_addr[1],
        'from' : from_addr[1], # Leave off the name and only return the address
        'from_name' : from_addr[0], # Leave off the name and only return the address
        'to_email': to_addr[1],
        'to' : to_addr[1], # Leave off the name and only return the address
        'to_name' : to_addr[0], # Leave off the name and only return the address
        'attachments': attachments,
    })

class MailManHandler(object):
    def saveAttachment(self, attr, lib=None, owner=None, group=None):
        if lib is None:
            lib = MediaLibrary(name="email", owner=owner, group=group)
            lib.save()
        upload_kind = validate_upload(attr)
        media = MediaItem(library=lib, name=attr.name, owner=owner, kind=upload_kind, newfile=attr)
        media.save()
        return media

    def on_message(self, muid, msg):
        pass


class MailMan(object):
    def __init__(self, host, username, password, timeout=60, markseen=False):
        self.handlers = []
        self.host = host
        self.username = username
        self.password = password
        self.timeout = timeout
        self.markseen = markseen
        self.con = None
        self.messages = []
        
    def connect(self):
        log("creating connection")
        self.con = imaplib.IMAP4_SSL(self.host)
        log("logging in")
        self.con.login(self.username, self.password)

    def disconnect(self):
        self.con.logout()

    def addHandler(self, handler):
        self.handlers.append(handler)

    def searchMail(self, unseen_only=True, search_from=None, search_subject=None, max_messages=25):
        """
        search mail and return list of mail uids
        """
        if not self.con:
            self.connect()
        self.con.select(readonly=True)
        search = "ALL"
        if unseen_only:
            search = "UNSEEN"
        search = [search]
        if search_subject:
            search.append('SUBJECT "{0}"'.format(search_subject))
        if search_from:
            search.append('FROM "{0}"'.format(search_from))
        search = "({0})".format(' '.join(search))
        result, uids = self.con.uid('search', None, search) # search and return uids instead
        uids = uids[0].split()
        log("{0}".format(uids))
        if len(uids) > max_messages:
            uids = uids[:max_messages]
        return uids

    def checkMail(self, unseen_only=True, search_from=None, search_subject=None, max_messages=25):
        self.messages = []
        if not self.con:
            self.connect()
        self.con.select(readonly=True)
        log("checking inbox for unread mail")
        #search = "ALL"
        # search = "SINCE {0} UNSEEN".format(datetime.now().strftime("%d-%m-%Y"))
        search = "ALL"
        if unseen_only:
            search = "UNSEEN"
        search = [search]
        if search_subject:
            search.append('SUBJECT "{0}"'.format(search_subject))
        if search_from:
            search.append('FROM "{0}"'.format(search_from))
        search = "({0})".format(' '.join(search))
        result, uids = self.con.uid('search', None, search) # search and return uids instead
        uids = uids[0].split()
        log("{0}".format(uids))
        if len(uids) > max_messages:
            uids = uids[:max_messages]
        for uid in uids:
            self.messages.append(self.fetchMessage(uid))
        log("check complete")
        messages = self.messages
        self.messages = []
        return messages

    def pushToHandlers(self, muid, msg):
        handled = False
        msg.id = muid
        for handler in self.handlers:
            if handler.on_message(muid, msg):
                handled = True
        if handled or self.markseen:
            self.markSeen(muid)

    def fetchMessage(self, muid):
        log("fetch message({0})".format(muid))
        self.con.select(readonly=True)
        results, data = self.con.uid('fetch', muid, "(RFC822)")
        if results != "OK":
            log("ERROR: fetchMessage({0}) failed with response({1})".format(muid, results))
            return False
        log("message fetched")
        msg = parse_raw_message(email.message_from_string(toString(data[0][1])))
        self.pushToHandlers(muid, msg)
        return msg

    def markSeen(self, muid):
        self.con.select(readonly=False)
        log("- marking '{0}' as seen".format(muid))
        typ, data = self.con.uid("STORE", muid,'+FLAGS','\Seen')
        log("- output '{0}' - '{1}'".format(typ, data))

    def markUnseen(self, muid):
        self.con.select(readonly=False)
        log("- marking '{0}' as unseen".format(muid))
        typ, data = self.con.uid("STORE", muid,'-FLAGS','\Seen')
        log("- output '{0}' - '{1}'".format(typ, data))

    def done(self):
        log("done")
        self.con.send("DONE\r\n")
        self.con.loop = False
        self.con.readline() #otherwise imaplib.idle() freaks out



# Generated by Django 4.1.4 on 2023-01-23 05:04

from django.db import migrations, models
import django.db.models.deletion
import rest.models


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0012_settings_settingsmetadata'),
        ('inbox', '0003_attachment_content_type'),
    ]

    operations = [
        migrations.CreateModel(
            name='MailTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(db_index=True, max_length=255)),
                ('kind', models.CharField(default='email', max_length=124)),
                ('template', models.TextField()),
                ('group', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='templates', to='account.group')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-29 19:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0004_group_modified_alter_group_created'),
    ]

    operations = [
        migrations.AddField(
            model_name='member',
            name='auth_code_expires',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
    ]





try:
    import cv2
    import numpy as np
except:
    cv2 = None

def hasCV():
    return cv2 is not None

def isCV2(or_better=False):
    # grab the OpenCV major version number
    major = int(cv2.__version__.split(".")[0])
    # check to see if we are using *at least* OpenCV 2
    if or_better:
        return major >= 2
    # otherwise we want to check for *strictly* OpenCV 2
    return major == 255

def detectShape(contour):
    shape = "unidentified"
    peri = cv2.arcLength(contour, True)
    approx = cv2.approxPolyDP(contour, 0.04 * peri, True)
    if len(approx) == 3:
        shape = "triangle"
    elif len(approx) == 4:
        (x, y, w, h) = cv2.boundingRect(approx)
        ar = w / float(h)
        shape = "square" if ar >= 0.95 and ar <= 1.05 else "rectangle"
    elif len(approx) == 5:
        shape = "pentagon"
    else:
        shape = "circle"
    return shape

def findLargestContour(contours):
    if len(contours) > 0:
        return sorted(contours, key=cv2.contourArea, reverse=True)[0]
    return []

def getBoundingBox(contour):
    rect = cv2.minAreaRect(contour)
    box = cv2.boxPoints(rect)
    return rect, np.int0(box)

def contrastContours(frame):
    if len(frame.shape) == 2:
        gray = frame
    else:
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    # compute the Scharr gradient magnitude representation of the images
    # in both the x and y direction using OpenCV 2.4
    ddepth = cv2.cv.CV_32F if isCV2() else cv2.CV_32F
    gradX = cv2.Sobel(gray, ddepth=ddepth, dx=1, dy=0, ksize=-1)
    gradY = cv2.Sobel(gray, ddepth=ddepth, dx=0, dy=1, ksize=-1)
    # subtract the y-gradient from the x-gradient
    gradient = cv2.subtract(gradX, gradY)
    gradient = cv2.convertScaleAbs(gradient)
    # blur and threshold the image
    blurred = cv2.blur(gradient, (9, 9))
    (_, thresh) = cv2.threshold(blurred, 225, 255, cv2.THRESH_BINARY)
    # construct a closing kernel and apply it to the thresholded image
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (21, 7))
    closed = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel)
    # perform a series of erosions and dilations
    closed = cv2.erode(closed, None, iterations=4)
    closed = cv2.dilate(closed, None, iterations=4)
    return closed

def findContours2(frame):
    closed = contrastContours(frame)
    # find the contours in the thresholded image
    cnts = cv2.findContours(closed.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    return grab_contours(cnts)

def lightContrast(frame):
    if len(frame.shape) == 2:
        gray = frame
    else:
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    blurred = cv2.GaussianBlur(gray, (5, 5), 0)
    return cv2.threshold(blurred, 60, 255, cv2.THRESH_BINARY)[1]

def gaussianContrast(frame, max_value=255, block_size=11, c=2):
    if len(frame.shape) == 2:
        gray = frame
    else:
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    return cv2.adaptiveThreshold(gray,max_value,cv2.ADAPTIVE_THRESH_GAUSSIAN_C,\
                cv2.THRESH_BINARY,block_size,c)

def findContoursEx(frame):
    # convert the resized image to grayscale, blur it slightly,
    # and threshold it
    if len(frame.shape) == 2:
        gray = frame
    else:
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    blurred = cv2.GaussianBlur(gray, (5, 5), 0)
    thresh = cv2.threshold(blurred, 60, 255, cv2.THRESH_BINARY)[1]
    cnts = cv2.findContours(thresh.copy(), cv2.RETR_EXTERNAL,
        cv2.CHAIN_APPROX_SIMPLE)
    return grab_contours(cnts)

def findContours(frame):
    # convert the resized image to grayscale, blur it slightly,
    # and threshold it
    if len(frame.shape) == 2:
        gray = frame
    else:
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    cnts = cv2.findContours(gray, cv2.RETR_EXTERNAL,
        cv2.CHAIN_APPROX_SIMPLE)
    return grab_contours(cnts)
    

def grab_contours(cnts):
    if len(cnts) == 2:
        cnts = cnts[0]
    elif len(cnts) == 3:
        cnts = cnts[1]
    else:
        raise Exception(("Contours tuple must have length 2 or 3, "
            "otherwise OpenCV changed their cv2.findContours return "
            "signature yet again. Refer to OpenCV's documentation "
            "in that case"))
    return cnts



from rest import decorators as rd
from rest.views import restPermissionDenied, restStatus
from account import models as am


@rd.url(r'^member$')
@rd.url(r'^member/$')
@rd.url(r'^member/(?P<pk>\d+)$')
@rd.login_required
def rest_on_member(request, pk=None):
    return am.Member.on_rest_request(request, pk)


@rd.url(r'^member/me$')
@rd.login_optional
def member_me_action(request):
    if not request.user.is_authenticated:
        return restPermissionDenied(request, "not authenticated")
    if request.method == "GET":
        # request.session['ws4redis:memberof'] = request.member.getGroupUUIDs()
        return request.member.on_rest_get(request)
    elif request.method == "POST":
        return request.member.on_rest_post(request)
    return restStatus(request, False, error="not supported")


@rd.url(r'^authtoken$')
@rd.url(r'^authtoken/(?P<pk>\d+)$')
@rd.login_required
def rest_on_authtoken(request, pk=None):
    return am.AuthToken.on_rest_request(request, pk)


@rd.url(r'^session$')
@rd.url(r'^session/(?P<pk>\d+)$')
@rd.login_required
def rest_on_session(request, pk=None):
    return am.AuthSession.on_rest_request(request, pk)




from django.forms import ModelForm
from django import forms
from .models import *
from .utils import *

class MediaLibraryForm(ModelForm):
    class Meta:
        model = MediaLibrary
        fields = ('name', 'description', 'allowed_kinds')

class MediaItemForm(ModelForm):
    class Meta:
        model = MediaItem
        fields = ('name', 'description')

    def __init__(self, *args, **kwargs):
        if 'library' in kwargs:
            self._library = kwargs.pop('library')
        if 'kind' in kwargs:
            self._kind = kwargs.pop('kind')
        if 'state' in kwargs:
            self._state = kwargs.pop('state')
        return super(MediaItemForm, self).__init__(*args, **kwargs)

    def save(self, commit=True):
        ret = super(MediaItemForm, self).save(commit=False)
        
        if hasattr(self, "_library"):
            ret.library = self._library
        if hasattr(self, "_kind"):
            ret.kind = self._kind
        if hasattr(self, "_state"):
            ret.state = self._state

        if commit:
            ret.save()
        return ret

class MediaItemUploadForm(MediaItemForm):
    file = forms.FileField()

    def clean_file(self):
        if not 'file' in self.files:
            return None
        self._upload_kind = validate_upload(self.files['file'])
        return self.files['file']

    def save(self, commit=True):
        ret = super(MediaItemUploadForm, self).save(commit=False)
        if 'file' in self.files:
            ret.kind = self._upload_kind
            ret.newfile = self.files['file']
        try:
            ret.owner
        except User.DoesNotExist:
            ret.owner = self.data['__request'].user
        
        if commit:
            ret.save()
        return ret

class MediaItemCanUploadForm(MediaItemUploadForm):
    file = forms.FileField(required=False)

class MediaItemNewLiveForm(MediaItemForm):
    def __init__(self, *args, **kwargs):
        self._url = kwargs.pop('url')
        if 'url_local' in kwargs:
            self._url_local = kwargs.pop('url_local')
        return super(MediaItemNewLiveForm, self).__init__(*args, **kwargs)

    def save(self, commit=True):
        ret = super(MediaItemNewLiveForm, self).save(commit=False)
        ret.kind = 'L'
        ret.owner = self.data['__request'].user
        ret.newurl = self._url
        if hasattr(self, '_url_local'):
            ret.newurl_local = self._url_local
        if commit:
            ret.save()
        return ret

class RenditionSetForm(ModelForm):
    class Meta:
        model = RenditionSet
        fields = ('name', )

class RenditionDefinitionForm(ModelForm):
    class Meta:
        model = RenditionDefinition
        fields = ('name', 'description', 'use', 'active',)

    preset_name = forms.CharField(max_length=256, required=True)
    
    def __init__(self, *args, **kwargs):
        ret = super(RenditionDefinitionForm, self).__init__(*args, **kwargs)
        try:
            preset = getattr(getattr(self, 'instance', None), 'preset', None)
            if preset and preset.name:
                self.initial['preset_name'] = preset.name
        except RenditionPreset.DoesNotExist:
            pass
    
    def clean_preset_name(self):
        preset = RenditionPreset.objects.filter(short_name = self.cleaned_data['preset_name'])
        if not preset.count() == 1:
            preset = RenditionPreset.objects.filter(name = self.cleaned_data['preset_name'])
        if not preset.count() == 1:
            raise ValidationError('Invalid preset name')

        return preset[0]
    
    def save(self, commit=True):
        rend = super(RenditionDefinitionForm, self).save(commit=False)
        rend.preset = self.cleaned_data['preset_name']
        if commit:
            rend.save()
        return rend
        
class CuePointForm(ModelForm):
    class Meta:
        model = CuePoint
        fields = ('start', 'end')
    name = forms.CharField(max_length=127, required=True)
    description = forms.CharField(required=False)

    def __init__(self, *args, **kwargs):
        self.item = kwargs.pop('item')
        ret = super(CuePointForm, self).__init__(*args, **kwargs)

    def clean_start(self):
        if 'start' in self.cleaned_data:
            return self.cleaned_data['start']
        if self.item.kind == 'L':
            orig = self.item.original()
            if not orig:
                return 0
            return orig.get_meta('duration', 0)
        else:
            return 0
    def clean(self):
        if 'end' in self.cleaned_data and self.cleaned_data['start'] > self.cleaned_data['end']:
            raise forms.ValidationError("Cue end must be greater than start")
        return self.cleaned_data

    def save(self, commit=True):
        cue = super(CuePointForm, self).save(commit=False)
        try:
            cue.item
        except MediaItem.DoesNotExist:
            cue.item = self.item

        try:
            cue.subitem
        except MediaItem.DoesNotExist:
            subitem = MediaItem(library=cue.item.library,
                owner=cue.item.owner,
                name=self.cleaned_data.get('name', None),
                description=self.cleaned_data.get('description', None),
                kind='S',
                state=100,
            )
            subitem.save()
            cue.subitem = subitem
        if commit:
            cue.save()
            cue.subitem.new_render()
        return cue



from django.forms import ModelForm, fields
from django.forms.widgets import ClearableFileInput
from django.core.files.uploadedfile import UploadedFile
from django.core import validators
from django.utils.encoding import smart_text

from .views import chunkUploadedFile

import re
from datetime import datetime, date

class ModelFormWithMeta(ModelForm):
    def __init__(self, *args, **kwargs):
        ret = super(ModelFormWithMeta, self).__init__(*args, **kwargs)
        if self.instance:
            rel = getattr(self.instance, self.Meta.meta_set)
            extension = getattr(self.Meta, 'meta_fields_extension', ())
            extended = getattr(self.Meta, 'meta_extended_fields', {})
            for f in self.Meta.meta_fields:
                keyval = f
                value_field = self.Meta.meta_fields[f]
                values = value_field.split('.', 1)
                if len(values) == 2:
                    keyval = values[0]
                    value_field = values[1]

                args = { getattr(self.Meta, 'meta_key', 'key'): keyval }
                m = rel.filter(**args)[:1]
                if len(m) > 0:
                    self.initial[f] = getattr(m[0], value_field, None)
                    for e in extension:
                        self.initial[f + "__" + e] = getattr(m[0], e, None)
                    for e in extended.get(f, []):
                        self.initial[extended[f][e]] = getattr(m[0], e, None)

        return ret

    def _save_meta(self):
        rel = getattr(self.instance, self.Meta.meta_set)
        extension = getattr(self.Meta, 'meta_fields_extension', ())
        extended = getattr(self.Meta, 'meta_extended_fields', {})
        for f in self.Meta.meta_fields:
            keyval = f
            value_field = self.Meta.meta_fields[f]
            values = value_field.split('.', 1)
            if len(values) == 2:
                keyval = values[0]
                value_field = values[1]

            meta_key = getattr(self.Meta, 'meta_key', 'key')
            m = rel.filter(**{ meta_key: keyval })[:1]

            # all meta attributes to be changed
            mattr = {}
            if f in self.data or self.cleaned_data.get(f):
                mattr[value_field] = self.cleaned_data[f]
            for e in extension:
                ef = f + "__" + e
                if ef in self.data:
                    mattr[e] = self.cleaned_data[ef]
            if f in extended:
                for e in extended[f]:
                    if e in self.data:
                        mattr[extended[f][e]] = self.data[e]

            if len(mattr) > 0:
                if len(m) == 0 and (f in self.data or self.cleaned_data.get(f)):
                    # new relationship
                    mattr[meta_key] = keyval
                    rel.create(**mattr)
                elif len(m) > 0:
                    # update existing
                    for f in mattr:
                        setattr(m[0], f, mattr[f])
                        m[0].save()

    def save(self, commit=True):
        def save_meta():
            self._save_meta_m2m()
            self._save_meta()

        instance = super(ModelFormWithMeta, self).save(commit=commit)
        if commit:
            save_meta()
        else:
            self._save_meta_m2m = self.save_m2m
            self.save_m2m = save_meta
        return instance

class ResumableFileInput(ClearableFileInput):
    def value_from_datadict(self, data, files, name):
        upload = super(ResumableFileInput, self).value_from_datadict(data, files, name)
        if not upload:
            if name + '.path' in data and name + '.name' in data and name + '.content_type' in data and name + '.size' in data:
                try:
                    f = open(data.get(name + '.path'))
                except OSError:
                    pass
                else:
                    upload = UploadedFile(file=f, name=data.get(name + '.name'), content_type=data.get(name + '.content_type'), size=data.get(name + '.size'))
        if not upload:
            if name + '_sessionid' in data:
                try:
                    upload = chunkUploadedFile(data.get('__request'), data.get(name + '_sessionid'))
                except OSError:
                    pass

        if upload:
            files[name] = upload
        return upload

class DateField(fields.DateField):
    def to_python(self, value):
        if value == None or value == '':
            return None
        elif type(value) in (int,float):
            return date.fromtimestamp(value)
        elif type(value) in (str,str) and re.match('^-?[0-9]+$', value):
            try:
                return date.fromtimestamp(int(value))
            except:
                pass
        if type(value) is datetime:
            return value.date()
        return super(fields.DateField, self).to_python(value)

class DateTimeField(fields.DateTimeField):
    def __init__(self, *args, **kwargs):
        ret = super(fields.DateTimeField, self).__init__(*args, **kwargs)
        
    def to_python(self, value):
        if value == None or value == '':
            return None
        elif type(value) in (int,float):
            return datetime.fromtimestamp(value)
        elif type(value) in (str,str) and re.match('^-?[0-9]+$', value):
            try:
                return datetime.fromtimestamp(int(value))
            except:
                pass
        return super(fields.DateTimeField, self).to_python(value)

class CommaSeparatedListField(fields.Field):
    def __init__(self, max_items=None, min_items=None, *args, **kwargs):
        self.max_items, self.min_items = max_items, min_items
        super(CommaSeparatedListField, self).__init__(*args, **kwargs)
        if min_items is not None:
            self.validators.append(validators.MinLengthValidator(min_items))
        if max_items is not None:
            self.validators.append(validators.MaxLengthValidator(max_items))
                
    def to_python(self, value):
        if value in validators.EMPTY_VALUES:
            return []
        
        if type(value) in (str, str):
            value = value.split(',')

        ret = []
        for t in value:
            t = t.strip()
            if t:
                ret.append(str(t))

        return ret



# Copyright (C) 2013 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command-line tools for authenticating via OAuth 2.0

Do the OAuth 2.0 Web Server dance for a command line application. Stores the
generated credentials in a common file that is used by other example apps in
the same directory.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'
__all__ = ['argparser', 'run_flow', 'run', 'message_if_missing']


import http.server
import argparse
import httplib2
import logging
import os
import socket
import sys
import webbrowser

from oauth2client import client
from oauth2client import file
from oauth2client import util

try:
  from urllib.parse import parse_qsl
except ImportError:
  from cgi import parse_qsl

_CLIENT_SECRETS_MESSAGE = """WARNING: Please configure OAuth 2.0

To make this sample run you will need to populate the client_secrets.json file
found at:

   %s

with information from the APIs Console <https://code.google.com/apis/console>.

"""

# run_parser is an ArgumentParser that contains command-line options expected
# by tools.run(). Pass it in as part of the 'parents' argument to your own
# ArgumentParser.
argparser = argparse.ArgumentParser(add_help=False)
argparser.add_argument('--auth_host_name', default='localhost',
                        help='Hostname when running a local web server.')
argparser.add_argument('--noauth_local_webserver', action='store_true',
                        default=False, help='Do not run a local web server.')
argparser.add_argument('--auth_host_port', default=[8080, 8090], type=int,
                        nargs='*', help='Port web server should listen on.')
argparser.add_argument('--logging_level', default='ERROR',
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR',
                                 'CRITICAL'],
                        help='Set the logging level of detail.')


class ClientRedirectServer(http.server.HTTPServer):
  """A server to handle OAuth 2.0 redirects back to localhost.

  Waits for a single request and parses the query parameters
  into query_params and then stops serving.
  """
  query_params = {}


class ClientRedirectHandler(http.server.BaseHTTPRequestHandler):
  """A handler for OAuth 2.0 redirects back to localhost.

  Waits for a single request and parses the query parameters
  into the servers query_params and then stops serving.
  """

  def do_GET(s):
    """Handle a GET request.

    Parses the query parameters and prints a message
    if the flow has completed. Note that we can't detect
    if an error occurred.
    """
    s.send_response(200)
    s.send_header("Content-type", "text/html")
    s.end_headers()
    query = s.path.split('?', 1)[-1]
    query = dict(parse_qsl(query))
    s.server.query_params = query
    s.wfile.write("<html><head><title>Authentication Status</title></head>")
    s.wfile.write("<body><p>The authentication flow has completed.</p>")
    s.wfile.write("</body></html>")

  def log_message(self, format, *args):
    """Do not log messages to stdout while running as command line program."""
    pass


@util.positional(3)
def run_flow(flow, storage, flags, http=None):
  """Core code for a command-line application.

  The run() function is called from your application and runs through all the
  steps to obtain credentials. It takes a Flow argument and attempts to open an
  authorization server page in the user's default web browser. The server asks
  the user to grant your application access to the user's data. If the user
  grants access, the run() function returns new credentials. The new credentials
  are also stored in the Storage argument, which updates the file associated
  with the Storage object.

  It presumes it is run from a command-line application and supports the
  following flags:

    --auth_host_name: Host name to use when running a local web server
      to handle redirects during OAuth authorization.
      (default: 'localhost')

    --auth_host_port: Port to use when running a local web server to handle
      redirects during OAuth authorization.;
      repeat this option to specify a list of values
      (default: '[8080, 8090]')
      (an integer)

    --[no]auth_local_webserver: Run a local web server to handle redirects
      during OAuth authorization.
      (default: 'true')

  The tools module defines an ArgumentParser the already contains the flag
  definitions that run() requires. You can pass that ArgumentParser to your
  ArgumentParser constructor:

    parser = argparse.ArgumentParser(description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[tools.run_parser])
    flags = parser.parse_args(argv)

  Args:
    flow: Flow, an OAuth 2.0 Flow to step through.
    storage: Storage, a Storage to store the credential in.
    flags: argparse.ArgumentParser, the command-line flags.
    http: An instance of httplib2.Http.request
         or something that acts like it.

  Returns:
    Credentials, the obtained credential.
  """
  logging.getLogger().setLevel(getattr(logging, flags.logging_level))
  if not flags.noauth_local_webserver:
    success = False
    port_number = 0
    for port in flags.auth_host_port:
      port_number = port
      try:
        httpd = ClientRedirectServer((flags.auth_host_name, port),
                                     ClientRedirectHandler)
      except socket.error as e:
        pass
      else:
        success = True
        break
    flags.noauth_local_webserver = not success
    if not success:
      print('Failed to start a local webserver listening on either port 8080')
      print('or port 9090. Please check your firewall settings and locally')
      print('running programs that may be blocking or using those ports.')
      print()
      print('Falling back to --noauth_local_webserver and continuing with', end=' ')
      print('authorization.')
      print()

  if not flags.noauth_local_webserver:
    oauth_callback = 'http://%s:%s/' % (flags.auth_host_name, port_number)
  else:
    oauth_callback = client.OOB_CALLBACK_URN
  flow.redirect_uri = oauth_callback
  authorize_url = flow.step1_get_authorize_url()

  if not flags.noauth_local_webserver:
    webbrowser.open(authorize_url, new=1, autoraise=True)
    print('Your browser has been opened to visit:')
    print()
    print('    ' + authorize_url)
    print()
    print('If your browser is on a different machine then exit and re-run this')
    print('application with the command-line parameter ')
    print()
    print('  --noauth_local_webserver')
    print()
  else:
    print('Go to the following link in your browser:')
    print()
    print('    ' + authorize_url)
    print()

  code = None
  if not flags.noauth_local_webserver:
    httpd.handle_request()
    if 'error' in httpd.query_params:
      sys.exit('Authentication request was rejected.')
    if 'code' in httpd.query_params:
      code = httpd.query_params['code']
    else:
      print('Failed to find "code" in the query parameters of the redirect.')
      sys.exit('Try running with --noauth_local_webserver.')
  else:
    code = input('Enter verification code: ').strip()

  try:
    credential = flow.step2_exchange(code, http=http)
  except client.FlowExchangeError as e:
    sys.exit('Authentication has failed: %s' % e)

  storage.put(credential)
  credential.set_store(storage)
  print('Authentication successful.')

  return credential


def message_if_missing(filename):
  """Helpful message to display if the CLIENT_SECRETS file is missing."""

  return _CLIENT_SECRETS_MESSAGE % filename

try:
  from .old_run import run
  from .old_run import FLAGS
except ImportError:
  def run(*args, **kwargs):
    raise NotImplementedError(
        'The gflags library must be installed to use tools.run(). '
        'Please install gflags or preferrably switch to using '
        'tools.run_flow().')



from rest import decorators as rd
from account.models.settings import Settings


@rd.url(r'^settings$')
@rd.url(r'^settings/$')
@rd.url(r'^settings/(?P<pk>\d+)$')
@rd.staff_required
def on_rest_settings(request, pk=None):
    return Settings.on_rest_request(request, pk)



from rest import settings
from django.apps import apps
from rest import helpers

from account.models.session import AuthSession
from rest.jwtoken import JWToken, JWT_KEY

# HACK FOR SameSite
from django.utils.cache import patch_vary_headers
from http.cookies import Morsel
Morsel._reserved['samesite'] = 'SameSite'

JWT_COOKIE_KEY = settings.get("JWT_COOKIE_KEY", "JWT_TOKEN")
JWT_ALLOW_COOKIE = settings.get("JWT_ALLOW_COOKIE", False)


class JWTokenMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        request.clear_jwt_cookie = False
        self.process_request(request)
        response = self.get_response(request)
        self.process_response(request, response)
        return response

    def process_request(self, request):
        request.jwt = None
        request.token = None
        request.token_bearer = None
        request.device_id = None
        request.auth_model = None
        request.member = None
        request.auth_session = None
        if not hasattr(request, "ip"):
            request.ip = helpers.getRemoteIP(request)
        token = request.META.get('HTTP_AUTHORIZATION', "").strip()
        if token.count(" ") != 1:
            self.process_cookie(request)
            return
        # helpers.log_error("token auth in progress", token)
        request.token_bearer, request.token = token.split(' ')
        if request.token_bearer.lower() == "bearer":
            self.process_jwt(request)
        elif request.token_bearer.lower() == "authtoken":
            self.process_authtoken(request)
        if request.member is not None:
            request.member.touchActivity()

    def process_cookie(self, request):
        if not JWT_ALLOW_COOKIE:
            return
        token = request.COOKIES.get(JWT_COOKIE_KEY, None)
        if token is not None:
            # helpers.log_error("cookie found")
            request.token = token
            request.token_bearer = "bearer"
            self.process_jwt(request)
            if request.token is None:
                # invalid JWT let us delete cookie
                # tells middleware to delete cookie in response
                request.clear_jwt_cookie = True

    def process_jwt(self, request):
        # this is JWT so let us authenticate
        token = JWToken(request.token)
        if token.payload is None:
            helpers.log_error("invalid jwt token", request.token)
            request.token = None
            request.token_bearer = None
            return
        User = apps.get_model("account", "User")
        user = User.objects.filter(pk=token.payload.user_id).last()
        if user is None:
            helpers.log_error("invalid jwt user", request.token)
            request.token = None
            request.token_bearer = None
            return
        member = user.getMember()
        if member.security_token is None:
            member.security_token = JWT_KEY
            member.save()
        # set the token key to the member security token, this lets us invalidate tokens
        token.key = member.security_token
        if not token.is_valid:
            # helpers.log_error("invalid jwt", request.token)
            request.token = None
            request.token_bearer = None
            return
        if not token.isExpired():
            helpers.log_error("expired jwt", request.token)
            request.token = None
            request.token_bearer = None
            return
        if not member.canLogin(request, False):
            helpers.log_error("user cannot login via jwt", request.token)
            request.token = None
            request.token_bearer = None
            return
        request.jwt = token
        request.user = user
        request.member = member
        request.signature = token.session_id
        request.device_id = token.payload.device_id
        request.auth_session = AuthSession.GetSession(request)

    def process_authtoken(self, request):
        AuthToken = apps.get_model("account", "AuthToken")
        atoken = AuthToken.objects.filter(token=request.token).last()
        if atoken is None:
            helpers.log_error("login failed with authtoken")
            return
        if not atoken.member.canLogin(request, False):
            helpers.log_error("user cannot login via authtoken", request.token)
            return 
        request.user = atoken.member.getUser()
        request.signature = atoken.signature
        request.member = atoken.member
        request.token_bearer = "auth_token"

    def process_response(self, request, response):
        if not JWT_ALLOW_COOKIE:
            return
        if request.token_bearer and request.token_bearer.lower() == "bearer":
            cookie_token = request.COOKIES.get(JWT_COOKIE_KEY, None)
            if cookie_token != request.token:
                # this is a new authed jwt token, store it in the cookie
                # patch_vary_headers(response, ('Cookie',))
                response.set_cookie(
                    key=JWT_COOKIE_KEY,
                    value=request.token,
                    secure=False,
                    httponly=False,
                    max_age=86400)
                response.cookies[JWT_COOKIE_KEY]["samesite"] = settings.get("SESSION_COOKIE_SAMESITE", "Lax")
        elif request.clear_jwt_cookie:
            response.delete_cookie(JWT_COOKIE_KEY)




# Generated by Django 4.0.2 on 2022-02-22 01:23

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('location', '__first__'),
    ]

    operations = [
        migrations.CreateModel(
            name='SessionLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('session_id', models.CharField(db_index=True, help_text='Django Session ID', max_length=127)),
                ('ip', models.CharField(blank=True, db_index=True, max_length=127, null=True)),
                ('user_agent', models.TextField(blank=True, help_text='Browser user agent', null=True)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True, help_text='When session started')),
                ('is_closed', models.BooleanField(blank=True, db_index=True, default=False)),
                ('location', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='sessions', to='location.geoip')),
                ('old_session', models.ForeignKey(blank=True, help_text='Previous session of same user agent', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='next_session', to='sessionlog.sessionlog')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='sessions', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from rest import helpers as rh
from rest import views as rv
from rest.log import getLogger
from rest import settings
from taskqueue.models import Task
from medialib.stores import s3store
from .models import Bounce, Complaint, Message, Attachment, Mailbox
from .utils import parsing as mailtils
import metrics

import requests
from objict import objict


logger = getLogger("inbox", filename="inbox.log")


def on_subscriptionconfirmation(request):
    logger.info("subcribed to", request.DATA.asDict())
    url = request.DATA.get("SubscribeURL", None)
    resp = requests.get(url)
    return rv.restStatus(request, True)


def on_bounce(request, msg):
    if not msg.bounce or not isinstance(msg.bounce.bouncedRecipients, list):
        logger.error("invalid bounce request")
        return rv.restStatus(request, True)
    for who in msg.bounce.bouncedRecipients:
        # kind, address, reason, reporter=None, code=None, source=None, source_ip=None, user=None
        metrics.metric("emails_bounced", category="email", min_granularity="hourly")
        Bounce.log(
            kind="email",
            address=who.emailAddress,
            reason=who.diagnosticCode,
            reporter=msg.bounce.reportingMTA,
            code=who.status,
            source=msg.mail.source,
            source_ip=msg.mail.sourceIp)
    return rv.restStatus(request, True)


def on_complaint(request, msg):
    if not msg.bounce or not isinstance(msg.bounce.bouncedRecipients, list):
        logger.error("invalid bounce request")
        return rv.restStatus(request, True)
    for who in msg.complaint.complainedRecipients:
        # kind, address, reason, reporter=None, code=None, source=None, source_ip=None, user=None
        metrics.metric("emails_complaints", category="email", min_granularity="hourly")
        Complaint.log(
            kind="email",
            address=who.emailAddress,
            reason=msg.complaint.complaintFeedbackType,
            user_agent=msg.complaint.userAgent,
            source=msg.mail.source,
            source_ip=msg.mail.sourceIp)
    return rv.restStatus(request, True)


def on_email(request, msg):
    """
    Email receiving can be configured in 2 ways.
     1. raw email via SNS which would have content field
     2. s3 bucket stored email which will have a receipt.action.bucketName
    """
    if msg.content is None and msg.receipt and msg.receipt.action:
        action = msg.receipt.action
        if action.type == "S3":
            return on_s3_email(request, msg, action.bucketName, action.objectKey)

    if msg.content is None:
        logger.error("message has no content", msg)
        return rv.restStatus(request, False)

    msg_data = mailtils.parseRawMessage(msg.content)
    return on_raw_email(request, msg, msg_data)


def on_s3_email(request, msg, bucket_name, object_key):
    msg_data = mailtils.parseRawMessage(s3store.getObjectContent(bucket_name, object_key))
    return on_raw_email(request, msg, msg_data)


def createMessage(to_email, msg_data):
    msg = Message(
        to_email=to_email,
        sent_at=msg_data.sent_at,
        subject=msg_data.subject,
        message=msg_data.message,
        html=msg_data.html,
        body=msg_data.body,
        to=msg_data.to,
        cc=msg_data.cc,
        from_email=msg_data.from_email,
        from_name=msg_data.from_name)
    msg.save()
    return msg


def on_raw_email(request, imsg, msg_data):
    to_email = imsg.receipt.recipients[0]
    # logger.info("parsed", msg_data)
    msg = createMessage(to_email, msg_data)
    metrics.metric("emails_received", category="email", min_granularity="hourly")

    attachments = []
    for msg_atch in msg_data.attachments:
        atch = Attachment(message=msg, name=msg_atch.name, content_type=msg_atch.content_type)
        if msg_atch.encoding == "base64":
            atch.saveMediaFile(msg_atch.payload, "media", msg_atch.name, is_base64=True)
        elif msg_atch.encoding == "quoted-printable":
            obj = mailtils.toFileObject(msg_atch)
            atch.saveMediaFile(obj, "media", msg_atch.name)
        atch.save()
        attachments.append(atch)
    # add the recipients mailbox
    addToMailbox(msg)
    # now lets check if we have more recipients
    for to_email in imsg.receipt.recipients[1:]:
        msg = createMessage(to_email, msg_data)
        for atch in attachments:
            # create a copy of the attachment, for the new msg
            atch.pk = None
            atch.message = msg
            atch.save()
        addToMailbox(msg)
    return rv.restStatus(request, True)


def addToMailbox(msg):
    # now lets find if a mailbox exists
    mailbox = Mailbox.objects.filter(email=msg.to_email).last()
    if mailbox is not None:
        Task.Publish(
            mailbox.tq_app, mailbox.tq_handler,
            data=dict(pk=msg.pk, to_email=msg.to_email, from_email=msg.from_email),
            channel=mailbox.tq_channel)


def on_notification(request):
    msg = objict.fromJSON(request.DATA.get("Message", ""))
    handler = SES_HANDLERS.get(msg.notificationType, None)
    if handler is None:
        logger.error(f"no handler for {msg.notificationType}")
        return rv.restStatus(request, False)
    try:
        return handler(request, msg)
    except Exception:
        logger.exception()
    return rv.restStatus(request, False)


def on_unknown(request):
    return rv.restStatus(request, False)


SES_HANDLERS = {
    "SubscriptionConfirmation": on_subscriptionconfirmation,
    "Bounce": on_bounce,
    "Complaint": on_complaint,
    "Notification": on_notification,
    "Received": on_email
}





from pushit.models import Product, Release
from rest import decorators as rd


@rd.urlPOST (r'^product$')
@rd.urlPOST (r'^product/(?P<product_id>\d+)$')
@rd.urlPOST (r'^product/uuid/(?P<uuid>\w+)$')
@rd.login_optional
def updateProduct(request, product_id=None, uuid=None):
    if not request.member:
        return restPermissionDenied(request)
    product = None
    if not product_id and not uuid:
        product = Product.createFromRequest(request, owner=request.member, group=request.group)
    elif product_id:
        product = Product.objects.filter(pk=product_id).last()
    elif uuid:
        product = Product.objects.filter(oid=uuid).last()

    if not product:
        return restStatus(request, False, error="unknown product")
    if product.owner != request.member or (product.group and not request.member.isMemberOf(product.group)):
        if not request.user.is_staff:
            return restPermissionDenied(request)
        product.saveFromRequest(request, owner=request.member)
    return restGet(request, product, **Product.getGraph("default"))


@rd.urlGET (r'^product/(?P<product_id>\d+)$')
@rd.urlGET (r'^product/uuid/(?P<uuid>\w+)$')
@rd.login_optional
def getProduct(request, product_id=None, uuid=None):
    product = None
    if product_id:
        product = Product.objects.filter(pk=product_id).last()
    elif uuid:
        product = Product.objects.filter(oid=uuid).last()
    else:
        return restNotFound(request)

    if not product:
        return restStatus(request, False, error="unknown product")
    if not product.is_public and not request.member:
        return restPermissionDenied(request, "not logged in")
    return product.restGet(request)


@rd.urlGET (r'^product$')
@rd.login_optional
def listProducts(request):
    if not request.member:
        return restPermissionDenied(request)
    if not request.member.is_staff:
        return restPermissionDenied(request)

    kind = request.DATA.get("kind")
    qset = Product.objects.filter(archived=False)
    if kind:
        qset = qset.filter(kind=kind)

    return restList(request, qset, **Product.getGraph("default"))


@rd.urlPOST (r'^release$')
@rd.urlPOST (r'^release/(?P<release_id>\d+)$')
@rd.login_optional
def updateRelease(request, release_id=None):
    if not request.member:
        return restPermissionDenied(request)

    if not release_id:
        auto_version = request.DATA.get("auto_version", False)
        prod_uuid = request.DATA.get(["product", "product_uuid"])
        product = None
        if prod_uuid:
            product = Product.objects.filter(oid=prod_uuid).last()
        else:
            prod_id = request.DATA.get("product_id")
            if prod_id:
                product = Product.objects.filter(pk=prod_id).last()
        if not product:
            return restStatus(request, False, error="product required")
        version_num = request.DATA.get("version_num", field_type=int)
        last_release = Release.objects.filter(product=product).order_by("-version_num").first()

        if not version_num:
            if last_release and auto_version:
                version_num = last_release.version_num + 1
            elif auto_version:
                version_num = 1
            else:
                return restStatus(request, False, error="no version info supplied, try auto_version=1")
        elif last_release and version_num <= last_release.version_num:
            return restStatus(request, False, error="version is not greater then last")

        release = Release.createFromRequest(request, product=product, owner=request.member, group=request.group, version_num=version_num)
    else:
        release = Release.objects.filter(pk=release_id).last()
        if not release:
            return restStatus(request, False, error="unknown release")
        if release.owner != request.member or (release.product.group and not request.member.isMemberOf(release.product.group)):
            if not request.user.is_staff:
                return restPermissionDenied(request)
        release.saveFromRequest(request, owner=request.member)
    if request.DATA.get("make_current"):
        release.makeCurrent()
    elif request.DATA.get("make_beta"):
        release.makeCurrent()
    return restGet(request, release, **Release.getGraph("default"))


@rd.urlGET (r'^release/(?P<release_id>\d+)$')
@rd.login_optional
def getRelease(request, release_id):
    release = Release.objects.filter(pk=release_id).last()
    if not release:
        return restStatus(request, False, error="unknown release")
    if not release.product.is_public and not request.member:
        return restPermissionDenied(request, "not logged in")
    return restGet(request, release, **Release.getGraph("default"))


def reportRestIssue(subject, message, perm="rest_errors", email_only=False):
    # notifyWithPermission(perm, subject, message=None, template=None, context=None, email_only=False)
    # notify email only
    Member.notifyWithPermission(perm, subject, message, email_only=email_only)


from django.db import models
from rest import settings

from rest.models import RestModel, MetaDataModel, MetaDataBase
from rest.fields import JSONField
from objict import nobjict
from rest import crypto
from ws4redis import client as redis
from account.models import Member
import importlib

import traceback

import time
from datetime import datetime, timedelta
NOT_FOUND = "-NOT-FOUNT-"

TASK_STATE_SCHEDULED = 0
TASK_STATE_STARTED = 1
TASK_STATE_RETRY = 2
TASK_STATE_COMPLETED = 10
TASK_STATE_FAILED = -1
TASK_STATE_CANCELED = -2

TASK_STATES = [
    (TASK_STATE_SCHEDULED, 'scheduled'),
    (TASK_STATE_STARTED, 'started'),
    (TASK_STATE_RETRY, 'retry_later'),
    (TASK_STATE_COMPLETED, 'completed'),
    (TASK_STATE_FAILED, 'failed'),
    (TASK_STATE_CANCELED, 'canceled')
]

def getAppHandler(app_name, fname):
    try:
        # module = __import__(app_name + '.tq', globals(), locals(), ['*'], 0)
        module = importlib.import_module(app_name + '.tq')

    except ImportError as err:
        return None

    if hasattr(getattr(module, fname, None), '__call__'):
        return getattr(module, fname)
    return None


class Task(models.Model, RestModel):
    """
    This is the state of a remote task.  This model is a backup store to a task that was scheduled
    via Task.Publish(data)
    TQ_SUBSCRIBE = ["tq_web_request", "tq_model_handler", "tq_app_handler"]

    tq_model_handler is a method on a django Model
    tq_app_handler is a method in a modules tq module... example (mymodule.tq.on_tq_test)

    Extends:
        models.Model
        RestModel
    """
    class RestMeta:
        DEFAULT_SORT = "-modified"
        POST_SAVE_FIELDS = ["action"]
        SEARCH_FIELDS = ["channel", "model", "fname", "data"]
        SEARCH_TERMS = [
            "channel", "model", "fname",
            "data", "reason", "runtime", "state"
        ]
        GRAPHS = {
            "list": {
                "fields":[
                    'id',
                    'created',
                    'modified',
                    'started_at',
                    'completed_at',
                    'stale_after',
                    'scheduled_for',
                    'cancel_requested',
                    'state',
                    'runtime',
                    'channel',
                    'model',
                    'fname',
                    ('truncate_data', 'data'),
                    'reason',
                    '_started',
                ],
                "extra": [("get_state_display", "state_display")],
            },
            "default": {
                "extra": [("get_state_display", "state_display")],
            },
            "detailed": {
                "extra": [("get_state_display", "state_display")],
            },
        }
    created = models.DateTimeField(db_index=True, auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    started_at = models.DateTimeField(default=None, null=True, blank=True)
    completed_at = models.DateTimeField(default=None, null=True, blank=True)
    stale_after = models.DateTimeField(default=None, null=True, blank=True)
    scheduled_for = models.DateTimeField(default=None, null=True, blank=True, db_index=True)
    cancel_requested = models.BooleanField(default=False, blank=True)
    state = models.IntegerField(db_index=True, default=TASK_STATE_SCHEDULED, choices=TASK_STATES)
    runtime = models.IntegerField(default=0)
    attempts = models.IntegerField(default=0)
    channel = models.CharField(max_length=200, db_index=True, default="tq_task")
    model = models.CharField(max_length=200, db_index=True)
    fname = models.CharField(max_length=200, default=None, null=True, blank=True)
    data = JSONField(default=None, null=True, blank=True)
    reason = models.CharField(max_length=255, default=None, null=True, blank=True)
    _started = 0

    @property
    def is_stale(self):
        if not self.stale_after:
            return False
        return self.stale_after <= datetime.now()

    @property
    def truncate_data(self):
        if self.data and self.data.data:
            temp = self.data
            temp.data = "...truncated..."
            return temp
        return self.data

    @property
    def created_age(self):
        return (datetime.now() - self.created).total_seconds()

    @property
    def modified_age(self):
        return (datetime.now() - self.modified).total_seconds()

    @property
    def current_runtime(self):
        if self.state in [10, -1]:
            return self.runtime
        if not self.started_at:
            return self.runtime
        return (datetime.now() - self.started_at).total_seconds()

    def set_action(self, value):
        if value == "retry_now":
            self.retry_now()
        elif value == "cancel":
            request = self.getActiveRequest()
            self.reason = "cancel request by {}".format(request.member.username)
            self.cancel(request.DATA.get("reason", "canceled by {}".format(request.member.username)))

    # def auditlog(self, action, message, request=None, path=None, component="taskqueue.Task"):
    #     # message, level=0, request=None, component=None, pkey=None, action=None, group=None, path=None, method=None
    #     # PersistentLog.log(message=message, level=1, action=action, request=request, component=component, pkey=self.pk, path=path, method=self.fname)
    #     raise Exception("auditlog called but is not supported")

    def log(self, text, kind="info"):
        TaskLog.Log(self, text, kind=kind)

    def log_exception(self, text, kind="exception"):
        self.log(str(text), kind=kind)
        self.log(traceback.format_exc(), "exception")

    def started(self):
        self.state = TASK_STATE_STARTED
        self.started_at = datetime.now()
        self._started = time.time()
        self.attempts += 1
        self.save()

    def retry_later(self, reason=None, from_now_secs=None):
        if reason:
            self.reason = reason
        if from_now_secs:
            # this will not run the task again until after this time has been hit
            self.scheduled_for = datetime.now() + timedelta(seconds=from_now_secs)
        self.state = TASK_STATE_RETRY
        self.cancel_requested = False
        self.save()

    def retry_now(self):
        self.state = TASK_STATE_SCHEDULED
        self.cancel_requested = False
        self.save()
        self.publish()

    def completed(self):
        self.completed_at = datetime.now()
        self.runtime = int(time.time() - self._started)
        self.state = TASK_STATE_COMPLETED
        self.save()

    def failed(self, reason=None):
        if reason and len(reason) > 250:
            reason = reason[:250]
        self.reason = reason
        self.state = TASK_STATE_FAILED
        self.notifyError()
        self.save()

    def notifyError(self, kind="Failure"):
        subject = "TaskQueue - {}".format(kind)
        handler = "unknown bg_handler"
        if self.data:
            handler = self.data.get("bg_handler")
        msg = "{}:{}<br>\n{}".format(self.model, handler, self.reason)
        # sms_msg = "{}\n{}".format(subject, msg)
        # Member.notifyWithPermission("taskqueue_alerts", subject, msg, sms_msg=sms_msg)
        metadata = {
            "server": settings.get("HOSTNAME", "unknown"),
            "task": self.pk,
            "reason": self.reason,
            "kind": kind,
            "app": self.model,
            "fname": self.fname,
            "channel": self.channel,
            "handler": handler
        }
        try:
            import incident
            incident.event_now(
                "taskqueue_errors", description=subject, details=msg, 
                level=3, metadata=metadata)
        except Exception as err:
            self.log(str(err), kind="error")

    def cancel(self, reason=None):
        self.cancel_requested = True
        self.save()
        out = nobjict()
        out.reason = reason
        out.pk = self.pk
        return redis.publish("tq_cancel", out)

    def publish(self, channel=None):
        out = nobjict()
        out.pk = self.pk
        out.model = self.model
        out.fname = self.fname
        out.data = self.data
        if not channel:
            channel = self.channel
        return redis.publish(channel, out)

    def getHandler(self):
        if "." not in self.model:
            return getAppHandler(self.model, self.fname)
        app, mname = self.model.split('.')
        model = self.restGetModel(app, mname)
        if not model:
            return None
        return getattr(model, self.fname, None)

    @classmethod
    def WebRequest(cls, url, data, fname="POST", stale_after_seconds=0):
        tdata = nobjict()
        tdata.url = url
        tdata.data = data
        task = cls(channel="tq_hook", model="tq_web_request", fname=fname, data=tdata)
        if stale_after_seconds:
            task.stale_after = datetime.now() + timedelta(seconds=stale_after_seconds)
        task.save()
        task.publish()
        return task

    @classmethod
    def EmailRequest(cls, address, data, filename=None, subject=None):
        tdata = nobjict()
        tdata.address = address
        tdata.filename = filename
        tdata.subject = subject
        tdata.data = data
        task = cls(channel="tq_hook", model="tq_email_request", data=tdata)
        task.save()
        task.publish()
        return task

    @classmethod
    def SMSRequest(cls, phone, data):
        tdata = nobjict()
        tdata.phone = phone
        tdata.data = data
        task = cls(channel="tq_hook", model="tq_sms_request", data=tdata)
        task.save()
        task.publish()
        return task

    @classmethod
    def SFTPRequest(cls, host, data, filename, username, password):
        tdata = nobjict()
        tdata.host = host
        tdata.filename = filename
        # TODO this should be more secure!
        # TODO support ssh keys?
        tdata.username = username
        tdata.password = password
        tdata.data = data
        task = cls(channel="tq_hook", model="tq_sftp_request", data=tdata)
        task.save()
        task.publish()
        return task
    
    @classmethod
    def S3Request(cls, bucket, data, folder, aws, secret, filename, when):
        tdata = nobjict()
        tdata.bucket = bucket
        tdata.filename = filename
        tdata.data = data
        tdata.folder = folder
        tdata.aws = aws
        tdata.secret = secret
        tdata.when = str(when)
        task = cls(channel="tq_hook", model="tq_s3_request", data=tdata)
        task.save()
        task.publish()
        return task

    @classmethod
    def PublishModelTask(cls, model_name, fname, data, stale_after_seconds=0, channel="tq_model_handler", scheduled_for=scheduled_for):
        return cls.Publish(cls, model_name, fname, data, stale_after_seconds, channel, scheduled_for)

    @classmethod
    def Publish(cls, app_name, fname, data=None, stale_after_seconds=0, channel="tq_app_handler", scheduled_for=None):
        # tq_handler will check for a function in the django app tq.py
        task = cls(model=app_name, fname=fname, data=data, channel=channel, scheduled_for=scheduled_for)
        if stale_after_seconds:
            task.stale_after = datetime.now() + timedelta(seconds=stale_after_seconds)
        if scheduled_for is not None:
            # this means we just save as a retry, and let it be scheduled for later
            task.state = TASK_STATE_RETRY
            task.save()
            return task
        task.save()
        task.publish(channel)
        return task

    @classmethod
    def FromEvent(cls, event):
        if not event.data or not event.data.pk:
            return None
        return cls.objects.filter(pk=event.data.pk).last()

    @classmethod
    def PublishTest(cls, count=1, sleep_time=20.0):
        for i in range(1, count+1):
            cls.Publish("taskqueue", "on_tq_test", {"published_at": time.time(), "index":i, "sleep_time": sleep_time})

    @classmethod
    def RestartEngine(cls):
        redis.publish("tq_restart", {})


class TaskWorkerClient:
    def __init__(self, hostname, **kwargs):
        self.hostname = hostname
        uid = crypto.randomString(16)
        self.recv_channel = f"tq:{uid}:{settings.HOSTNAME}"
        self.send_channel = f"tq:host:{hostname}"
            
        self.client = None
        self.pubsub = None

    def connect(self):
        if self.client is None:
            self.client = redis.getRedisClient()
            self.pubsub = self.client.pubsub()
            self.pubsub.subscribe(self.recv_channel)
            self.pubsub.get_message(timeout=2.0)  # wait for subscription

    def close(self):
        if self.client is not None:
            self.pubsub.unsubscribe(self.recv_channel)
            self.client = None
            self.pubsub = None

    def send(self, data):
        self.connect()
        data["hostname"] = settings.HOSTNAME
        data["response_channel"] = self.recv_channel
        redis.publish(
            self.send_channel, 
            data, self.client)

    def recv(self, timeout=5.0):
        self.connect()
        msg = self.pubsub.get_message(timeout=timeout)
        if msg is None:
            raise Exception("no response")
        msg = nobjict.fromdict(msg)
        if msg.data:
            msg.data = nobjict.fromJSON(msg.data)
        return msg.data
            
    def ping(self):
        start = time.perf_counter()
        self.send(dict(action="ping"))
        try:
            response = self.recv()
        except Exception as err:
            response = nobjict(action="error", error=str(err))
        response.time = time.perf_counter() - start
        return response.action == "pong", response

    def get_stats(self):
        return self.send_and_recv("get_stats")

    def restart(self):
        return self.send_and_recv("restart")

    def send_and_recv(self, action):
        start = time.perf_counter()
        self.send(dict(action=action))
        try:
            response = self.recv()
        except Exception as err:
            response = nobjict(action="error", error=str(err))
        response.time = time.perf_counter() - start
        return response

    @classmethod
    def PING(cls, hostname):
        client = TaskWorkerClient(hostname)
        client.connect()
        status, msg = client.ping()
        client.close()
        return status, msg

    @classmethod
    def GET_STATS(cls, hostname):
        client = TaskWorkerClient(hostname)
        client.connect()
        msg = client.get_stats()
        client.close()
        return msg

    @classmethod
    def GET_ONLINE(cls):
        return [h.decode() for h in redis.smembers("tq:host:online")]


class TaskLog(models.Model, RestModel):
    created = models.DateTimeField(db_index=True, auto_now_add=True)
    task = models.ForeignKey(Task, related_name="logs", on_delete=models.CASCADE)
    kind = models.CharField(max_length=64, default=None, null=True, blank=True)
    text = models.TextField()

    @classmethod
    def Log(cls, task, text, kind="info"):
        log = cls(task=task, text=text, kind=kind)
        log.save()
        return log


class TaskHook(models.Model, RestModel, MetaDataModel):
    """
    This does nothing on its own.  It simply allows for the defining of task hooks
    that can be used during execution of ones own task logic.  The task logic would
    look up the group and channel (just a unique identifier for when the hook is caught)

    use properties to add extra metadata for this hook.
        - group     group this hook belongs to
        - kind      HTTP_POST, EMAIL, SMS, SFTP
        - data_format    file format, csv, json, pdf
        - endpoint  web address, email, sms or other
        - channel   a unique keyword to use to search when looking for the hook
        - model     django app.Model string to use to search when looking for
    """
    class RestMeta:
        CAN_DELETE = True
        DEFAULT_SORT = "-modified"
        SEARCH_FIELDS = ["channel", "model", "endpoint"]
        SEARCH_TERMS = [
            "channel", "model", "fname",
            "data", "reason", "runtime", "state"
        ]
        GRAPHS = {
            "list": {
                "extra":["metadata"],
                "graphs": {
                    "group":"basic"
                }
            },
            "default": {
                "extra":["metadata"],
                "graphs": {
                    "group":"basic"
                }
            },
            "detailed": {
                "graphs": {
                    "self":"default"
                }
            },
        }

    created = models.DateTimeField(db_index=True, auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    group = models.ForeignKey("account.Group", blank=True, null=True, default=None, related_name="+", on_delete=models.CASCADE)
    kind = models.CharField(max_length=200, blank=True, null=True)
    data_format = models.CharField(max_length=32, default="json")
    endpoint = models.CharField(max_length=200, blank=True, null=True)
    state = models.IntegerField(default=1, choices=[(0, 'inactive'), (1, 'active')], db_index=True)

    # channel is just a unique keyword to check for hooks
    channel = models.CharField(max_length=200, db_index=True, default="tq_hook")
    # if this only gets fired for a particular model
    model = models.CharField(max_length=200, blank=True, null=True, db_index=True)

    def trigger(self, data, when=None):
        task = None
        if self.kind == "HTTP_POST":
            # url, data, fname="POST", stale_after_seconds=0
            task = Task.WebRequest(self.endpoint, data)
        elif self.kind == "HTTP_GET":
            task = Task.WebRequest(self.endpoint, data, fname='GET')
        elif self.kind == "EMAIL":
            task = Task.EmailRequest(
                self.endpoint, data,
                self.getProperty("filename", "{date.month:02d}{date.day:02d}{date.year}." + self.data_format),
                self.getProperty("subject", "{date.month:02d}{date.day:02d}{date.year}"))
        elif self.kind == "SFTP":
            task = Task.SFTPRequest(
                self.endpoint, data,
                self.getProperty("filename", "{date.month:02d}{date.day:02d}{date.year}." + self.data_format),
                self.getProperty("username"),
                self.getProperty("password"))
        elif self.kind == "SMS":
            task = Task.SMSRequest(self.endpoint, data)
        elif self.kind == "S3":
            task = Task.S3Request(
                self.getProperty("bucket"), data,
                self.getProperty("folder", None),
                self.getProperty("aws", None),
                self.getProperty("secret", None),
                self.getProperty("filename", "{date.month:02d}{date.day:02d}{date.year}." + self.data_format),
                when)
        else:
            print("unknown hook kind: {}".format(self.kind))
        return task


class TaskHookMetaData(MetaDataBase):
    parent = models.ForeignKey(TaskHook, related_name="properties", on_delete=models.CASCADE)






from rest.decorators import periodic
from incident.models import Event, ServerOssecAlert, Incident
from datetime import datetime, timedelta
from rest import log

logger = log.getLogger("incident", filename="incident.log")


@periodic(minute=50, hour=8)
def run_cleanup(force=False, verbose=False, now=None):
    stale = datetime.now() - timedelta(days=90)
    # delete all ossec alerts older then 90 days
    count = ServerOssecAlert.objects.filter(created__lte=stale).delete()
    if count:
        logger.info(f"deleted {count} old ServerOssecAlert")
    # delete all events older then 90 days
    count = Event.objects.filter(created__lte=stale).delete()
    if count:
        logger.info(f"deleted {count} old Events")



import requests
import time
REQ_TIMEOUT = 15.0
from objict import objict


def POST(task, attempts=3, timeout=REQ_TIMEOUT, headers=None):
    while attempts:
        try:
            started = time.time()
            attempts -= 1
            data = task.data.data
            resp = httpPOST(task.data.url, data, isinstance(data, (dict, list)), timeout=timeout, headers=headers)
            task.log(resp.text)
            if resp.status_code == 200:
                return resp
        except requests.Timeout:
            task.log("timed out after {}s".format(time.time()-started), kind="error")
        except Exception as err:
            task.log_exception(err)
    return None

def GET(task, attempts=3, timeout=REQ_TIMEOUT):
    while attempts:
        try:
            started = time.time()
            attempts -= 1
            resp = httpGET(task.data.url, params=task.data.data, timeout=timeout)
            task.log(resp.text)
            if resp.status_code == 200:
                return resp
        except requests.Timeout:
            task.log("timed out after {}s".format(time.time()-started), action="error")
        except Exception as err:
            task.log_exception(err)
    return None


def httpGET(url, params=None, session=None, timeout=REQ_TIMEOUT):
    headers = {'Accept': 'application/json'}
    if session:
        res = session.get(url, headers=headers, params=params, timeout=timeout)
    else:
        res = requests.get(url, headers=headers, params=params, timeout=timeout)
    return res


def httpPOST(url, data, post_json=False, session=None, timeout=REQ_TIMEOUT, headers=None):
    if not headers or type(headers) is not dict:
        headers = {'Accept': 'application/json'}
    if 'Accept' not in headers:
        headers['Accept'] = 'application/json'
    if post_json:
        data = objict.fromdict(data).toJSON()
        headers['Content-type'] = 'application/json'
    if post_json:
        if session:
            res = session.post(url, json=data, headers=headers, timeout=timeout, verify=False)
        else:
            res = requests.post(url, json=data, headers=headers, timeout=timeout, verify=False)
    else:
        if session:
            res = session.post(url, data=data, headers=headers, timeout=timeout, verify=False)
        else:
            res = requests.post(url, data=data, headers=headers, timeout=timeout, verify=False)
    return res



from django.db import models
from django.db.models import Q
from django.contrib.auth.models import AbstractUser
from django.contrib.auth import login as auth_login, logout as auth_logout, authenticate
from django.contrib.auth.hashers import check_password
from django.contrib.sessions.models import Session

from datetime import datetime, timedelta
import time
import re
import hashlib
import uuid

from auditlog.models import PersistentLog
from sessionlog.models import SessionLog

from location.models import GeoIP
from rest import RemoteEvents
from rest.models import RestModel, MetaDataModel, MetaDataBase, RestValidationError, PermissionDeniedException
from rest.fields import FormattedField
from rest import helpers as rh
from rest import mail as rest_mail
from rest import crypto
from rest import settings
import incident
try:
    import pyotp
except Exception:
    pyotp = None


# we need bigger usernames, in particular for RemoteMember
AbstractUser._meta.get_field('username').max_length = 128
AUTH_CODE_EXPIRES_SECS = settings.get("AUTH_CODE_EXPIRES_SECS", 1440)


class User(AbstractUser, RestModel):
    class Meta:
        db_table = 'auth_user'

    class RestMeta:
        NO_SHOW_FIELDS = ["password"]

    def getMember(self):
        return Member.getByUser(self)

    def log(self, action, message, request=None, group=None, path=None, method=None, level=0):
        # message, level=0, request=None, component=None, pkey=None, action=None, group=None, path=None, method=None
        component = "account.Member"
        pkey = self.id
        PersistentLog.log(message=message, level=level, action=action, request=request, component=component, pkey=pkey, group=group, path=path, method=method)

    def getGroups(self):
        Group = RestModel.getModel("account", "Group")
        q = Q(memberships__member=self, memberships__state__gte=-10)
        q = q | Q(parent__memberships__member=self, parent__memberships__state__gte=-10)
        q = q | Q(parent__parent__memberships__member=self, parent__parent__memberships__state__gte=-10)
        return Group.objects.filter(q).distinct()
        # return Group.objects.filter(memberships__member=self, memberships__state__gte=-10).distinct()

    def getGroupIDs(self):
        return list(self.getGroups().values_list("pk", flat=True))

    def getGroupUUIDs(self):
        return list(self.getGroups().values_list("uuid", flat=True))


class Member(User, RestModel, MetaDataModel):
    class RestMeta:
        NO_SHOW_FIELDS = ["password", "auth_code", "security_token"]
        SEARCH_FIELDS = ["username", "email", "first_name", "last_name", "display_name"]
        VIEW_PERMS = ["view_members", "manage_members", "manage_users", "owner"]
        SAVE_PERMS = ["invite_members", "manage_members", "manage_users", "owner"]
        LIST_DEFAULT_FILTERS = {
            "is_active": True
        }
        NULLS_LAST = ["last_activity"]
        # note the | will check collection parameter...
        #   trailing "." will check if the collection has the key set to true
        SEARCH_TERMS = [
            "username", "email",
            "first_name", "last_name",
            "last_activity#datetime", "date_joined#datetime",
            "is_staff",
            ("notify_via", "properties|notify_via"),
            ("phone", "phone_number"),
            ("perms", "properties|permissions.")]
        UNIQUE_LOOKUP = ["username", "email"]
        METADATA_FIELD_PROPERTIES = settings.USER_METADATA_PROPERTIES
        GRAPHS = {
            "basic": {
                "fields": [
                    'id',
                    ('get_full_name', 'full_name'),
                    'first_name',
                    'last_name',
                    'display_name',
                    'initials',
                    'username',
                    'email',
                    'phone_number',
                    'last_login',
                    'last_activity',
                    'avatar'
                ]
            },
            "default": {
                "fields": [
                    'id',
                    'uuid',
                    'display_name',
                    ('get_full_name', 'full_name'),
                    'first_name',
                    'last_name',
                    'initials',
                    'username',
                    'email',
                    'phone_number',
                    'is_online',
                    'is_active',
                    'is_blocked',
                    'is_staff',
                    'is_superuser',
                    'requires_topt',
                    'last_login',
                    'last_activity',
                    ('date_joined', 'created'),
                    ("hasLoggedIn", "has_logged_in"),
                    'avatar',
                    'has_topt',
                    'auth_token'
                ],
                "extra": ["metadata", "password_expires_in"],
            },
        }

    uuid = models.CharField(db_index=True, max_length=64, blank=True, default="")
    modified = models.DateTimeField(auto_now=True)

    phone_number = FormattedField(format=FormattedField.PHONE, default=None, null=True, max_length=64, db_index=True)
    display_name = models.CharField(max_length=64, blank=True, null=True, default=None)
    picture = models.ForeignKey("medialib.MediaItem", blank=True, null=True, help_text="Profile picture", related_name='+', on_delete=models.CASCADE)

    # we use this token to allow us to invalidate JWT tokens 
    security_token = models.CharField(max_length=64, blank=True, null=True, default=None, db_index=True)
    auth_code = models.CharField(max_length=64, blank=True, null=True, default=None, db_index=True)
    auth_code_expires = models.DateTimeField(blank=True, null=True, default=None)

    password_changed = models.DateTimeField(blank=True, null=True, default=None)
    last_activity = models.DateTimeField(blank=True, null=True, default=None)
    requires_topt = models.BooleanField(blank=True, null=True, default=False)

    def __str__(self):
        return self.username

    def getUser(self):
        return self.user_ptr

    def getGroup(self, group_id, include_parents=True):
        # now we want to verify we are a member of this group or a parent group
        if isinstance(group_id, str) and not group_id.isdigit():
            return None
        Group = RestModel.getModel("account", "Group")
        if self.hasPerm("view_all_groups"):
            return Group.objects.filter(pk=group_id).first()
        ms = self.getMembershipFor(group_id, include_parents)
        if ms is not None:
            if ms.group.id != group_id:
                # this means we got a parent ms, so lets turn the group itself
                return Group.objects.filter(pk=group_id).first()
            else:
                return ms.group
        return None

    @property
    def initials(self):
        if self.first_name and self.last_name:
            return "{}{}".format(self.first_name[0], self.last_name[0])
        return None

    @property
    def full_name(self):
        return self.get_full_name()

    @property
    def is_online(self):
        return self.getActiveConnections()

    @property
    def is_blocked(self):
        when = RemoteEvents.hget("users:blocked:username", self.username)
        if not when:
            return False
        # check if still blocked
        now = time.time()
        if now - float(when) > settings.LOCK_TIME:
            self.unblock()
            return False
        return True

    @property
    def default_auth_token(self):
        return AuthToken.objects.filter(member=self, role="default").last()

    @property
    def auth_token(self):
        token = AuthToken.objects.filter(member=self, role="default").last()
        if token:
            return token.secure_token
        return None

    @property
    def has_topt(self):
        token = self.getSecretProperty("totp_token", default=None)
        return token is not None

    @property
    def password_expires_in(self):
        # this is returned in hours
        if self.password_changed is None:
            self.password_changed = datetime.now()
            self.save()
        days = (datetime.now() - self.password_changed).days
        return settings.PASSWORD_EXPIRES_DAYS - days

    @property
    def avatar(self):
        if self.picture is not None:
            return self.picture.thumbnail_url()
        return self.getProperty("google.picture", None)

    def set_avatar(self, value):
        self.saveMediaFile(value, "picture", None, True)

    def hasLoggedIn(self):
        if not self.last_login or not self.date_joined:
            return False

        if (self.last_login - self.date_joined).total_seconds() > 2:
            return self.has_usable_password()
        return False

    def recordFailedLogin(self, request):
        c = RemoteEvents.hincrby("users:failed:username", self.username, 1)
        if c >= settings.LOCK_PASSWORD_ATTEMPTS:
            self.block("multiple incorrect password attempts", request=request)
        else:
            self.reportIncident(
                "account", f"incorrect password for {self.username}", level=8,
                error_code=401,
                request=request)
        c = RemoteEvents.hincrby("users:failed:ip", request.ip, 1)

    def recordSuccessLogin(self, request):
        self.last_login = datetime.now()
        self.save()
        RemoteEvents.hdel("users:failed:username", self.username)
        RemoteEvents.hdel("users:failed:ip", request.ip)

    def hasPasswordExpired(self):
        now = datetime.now()
        if self.password_changed is None:
            self.password_changed = now
            self.save()
        return now - self.password_changed > timedelta(days=settings.PASSWORD_EXPIRES_DAYS)

    def login(self, password=None, request=None, use_jwt=False):
        if not self.is_active or self.is_blocked:
            return False
        # can force login
        if not request:
            request = rh.getActiveRequest()
        if password:
            if not self.checkPassword(password):
                # invalid password
                self.recordFailedLogin(request)
                return False
            else:
                self.recordSuccessLogin(request)
        if use_jwt:
            self.recordSuccessLogin(request)
            self.locateByIP(request.ip)
            return True
        self.user_ptr.backend = 'django.contrib.auth.backends.ModelBackend'
        auth_login(request, self.user_ptr)
        self.locateByIP(request.ip)
        return True

    def canLogin(self, request=None, throw_exception=True):
        if not self.is_active:
            self.log("login_blocked", F"account is not active {self.username}", request, method="login", level=31)
            if throw_exception:
                raise PermissionDeniedException("Account disabled", 410)
            return False
        if self.is_blocked:
            self.log("login_blocked", F"account is locked out {self.username}", request, method="login", level=31)
            if throw_exception:
                raise PermissionDeniedException("Account locked out", 411)
            return False
        if self.hasPasswordExpired():
            self.log("login_blocked", "password has expired", request, method="login", level=31)
            if throw_exception:
                raise PermissionDeniedException("password expired", 412)
            return False
        return True

    def touchActivity(self, force=False, last_login=False):
        is_dirty = False
        update_last_activity = datetime.now() - timedelta(minutes=5)
        if force or not self.last_activity or (self.last_activity and self.last_activity < update_last_activity):
            self.last_activity = datetime.now()
            is_dirty = True
        if last_login:
            update_last_login = datetime.now() - timedelta(hours=1)
            if not self.last_login or (self.last_login and self.last_login < update_last_login):
                self.last_login = datetime.now()
                self.last_activity = self.last_login
                is_dirty = True
        if is_dirty:
            # update the current device if one exists
            request = self.getActiveRequest()
            if request is not None and request.device_id:
                device = self.devices.filter(uuid=request.device_id).last()
                if device is not None:
                    device.touch(request.ip)
            self.save()

    def addPermission(self, perm):
        return self.setProperty(perm, 1, "permissions")

    def removePermission(self, perm):
        return self.setProperty(perm, None, "permissions")

    def hasPermission(self, perm, ignore_su=False):
        return self.hasPerm(perm, ignore_su)

    def hasPerm(self, perm, ignore_su=False):
        if not self.is_active:
            return False
        if self.is_superuser and not ignore_su:
            return True
        if isinstance(perm, list):
            for i in perm:
                if self.hasPerm(i, ignore_su):
                    return True
            return False
        return self.getProperty(perm, 0, "permissions", bool)

    def hasGroupPerm(self, group, perm):
        if group is None:
            return False
        ms = self.getMembershipFor(group)
        if ms is None or ms.state < -10:
            return False
        return ms.hasPerm(perm)

    def checkPassword(self, raw_password):
        """
        Return a boolean of whether the raw_password was correct. Handles
        hashing formats behind the scenes.
        """
        def setter(raw_password):
            super(Member, self).set_password(raw_password)
            # Password hash upgrades shouldn't be considered password changes.
            self._password = None
            self.save(update_fields=["password"])
        return check_password(raw_password, self.password, setter)

    def _setPassword(self, value):
        super(Member, self).set_password(value)
        self.save()

    def setPassword(self, value, skip_history=False):
        """
        if (this.length > 6) score++;
        if (this.length > 11) score++;
        if (this.length > 15) score++;

        //if this has both lower and uppercase characters give 1 point
        if ( ( this.match(/[a-z]/) ) && ( this.match(/[A-Z]/) ) ) score++;

        //if this has at least one number give 1 point
        if (this.match(/\d+/)) score++;

        //if this has at least one special caracther give 1 point
        if ( this.match(/[!@#$%^&*.]/) ) score++;

        """
        score = 0
        score += 1 if len(value) > 6 else 0
        score += 1 if len(value) > 8 else 0
        score += 2 if len(value) > 15 else 0
        score += 1 if re.match(r'.*[a-z]', value) and re.match(r'.*[A-Z]', value) else 0
        score += 1 if re.match(r'.*\d+', value) else 0
        score += 1 if re.match(r'.*[!@#$%^&*.]', value) else 0
        if score < 3:
            self.log("password_error", "password is weak or duplicate", method="password_change", level=21)
            raise RestValidationError("password is weak or duplicate")
        if not skip_history and settings.PASSWORD_HISTORY:
            hashed_password = PasswordHistory.HashPassword(value)
            # this means you can never reuse the same password
            if self.password_history.filter(password=hashed_password).count():
                self.log("password_error", "password already used", method="password_change", level=21)
                raise RestValidationError("password already used")
            if not self.id:
                self.save()
            PasswordHistory(owner=self, password=hashed_password).save()
        # self.set_password(value)
        self._setPassword(value)
        self.password_changed = datetime.now()
        self.save()
        self.log("password_changed", "password changed", method="password_change", level=21)
        return True

    def sendEvent(self, name, message, custom=None):
        if not custom:
            custom = {}
        custom["pk"] = self.id
        custom["member_id"] = self.id
        RemoteEvents.sendToUser(
            self,
            name,
            message=message,
            custom=custom)

    def sendChangeEvent(self, component, component_id, name="user.change", custom=None):
        if not custom:
            custom = {}
        custom["pk"] = self.id
        custom["member_id"] = self.id
        RemoteEvents.sendToUser(
            self,
            name,
            component=component,
            component_id=component_id,
            custom=custom)

    def isMemberOf(self, group, include_parents=True):
        ms = self.getMembershipFor(group, include_parents)
        return ms is not None

    def getMembershipFor(self, group, include_parents=True):
        # django should auto filter by group_id if int
        if group is None:
            return None
        if include_parents:
            if isinstance(group, (int, str)):
                Group = RestModel.getModel("account", "Group")
                group = Group.objects.get(pk=group)
            return group.getMembership(self, True)
            # now check by children
        return self.memberships.filter(group=group).first()

    def disable(self, by, reason="", notify=True):
        self.is_active = False
        self.save()
        self.memberships.update(state=-100)

        [s.delete() for s in Session.objects.all() if s.get_decoded().get('_auth_user_id') == self.pk]
        # notify account disabled
        subject = "MEMBER {} DISABLED BY {}".format(self, by)
        accounts = []
        body = "{}<br>\ndisabled from: ".format(reason)
        for m in self.memberships.all():
            accounts.append(m.group.name)
        body += "<br>\n".join(accounts)
        self.log("disabled", "account({}) disabled by {}, {}".format(self.username, by.username, reason), method="disabled", level=35)
        if notify:
            Member.notifyWithPermission("user_audit", subject, message=body, email_only=True)

    def enable(self, by, memberships=None):
        if not self.is_active:
            self.is_active = True
            self.save()
        # notify account disabled
        subject = "MEMBER {} RE-ENABLED BY {}".format(self, by)
        body = "enabled for: "
        if memberships is None:
            memberships = self.memberships.all()

        accounts = []
        for m in memberships:
            m.state = 10
            m.save()
            accounts.append(m.group.name)
        body += "<br>\n".join(accounts)
        self.log("enabled", "account enabled by {}".format(by.username), method="enabled", level=35)
        Member.notifyWithPermission("user_audit", subject, message=body, email_only=True)

    def locateByIP(self, ip):
        loc = GeoIP.get(ip)
        if loc is not None:
            self.setProperty("last_ip", ip)
            self.setProperty("city", loc.city, "location")
            self.setProperty("state", loc.state, "location")
            self.setProperty("country", loc.country, "location")
            self.setProperty("lat", loc.lat, "location")
            self.setProperty("lng", loc.lng, "location")

    def sendSMS(self, msg):
        from telephony.models import SMS
        phone = self.phone_number
        if not phone or len(phone) < 7:
            return False
        SMS.send(phone, msg)
        return True

    def sendEmail(self, subject, body, attachments=[], do_async=True, template=None, context=None):
        default_domain = self.getProperty("default_domain", settings.EMAIL_DEFAULT_DOMAIN)
        from_email = rh.getFromEmailForHost(default_domain)
        NotificationRecord = RestModel.getModel("account", "NotificationRecord")
        NotificationRecord.notify(
            [self], subject=subject, message=body,
            template=template, context=context,
            from_email=from_email,
            email_only=True, attachments=attachments)

    def sendInvite(self, subject, group=None, url=None, msg=None, **kwargs):
        if url is None:
            url = settings.DEFAULT_LOGIN_URL
        if url[:4] != "http":
            url = f"https://{url}"
        from_domain = rh.parseHostName(url, True)
        from_email = rh.getFromEmailForHost(from_domain)
        self.member.setProperty("default_domain", from_domain)
        self.member.setProperty("default_from_email", from_email)
        context = rh.getContext(
            self.getActiveRequest(),
            member=self,
            group=group,
            url=url,
            msg=msg, **kwargs)
        rest_mail.send(
            self.email, subject, from_email=from_email,
            template=settings.get("EMAIL_TEMPLATE_INVITE", "email/invite.html"),
            context=context)

    def getActiveSessions(self):
        return SessionLog.objects.filter(user__id=self.pk, is_closed=False)

    def getSessionCount(self):
        return self.getActiveSessions().count()

    def logout(self, request=None, all_sessions=False, older_then=None):
        if not request:
            request = rh.getActiveRequest()

        if request and request.member == self:
            self.log("logout", "user logged out", request, method="logout", level=18)
            auth_logout(request)
            if "member_id" in request.session:
                del request.session["member_id"]
            if "_auth_user_id" in request.session:
                del request.session["_auth_user_id"]

        # else:
        #     qset = self.getActiveSessions()
        #     if older_then:
        #         age = datetime.now() - timedelta(days=older_then)
        #         qset = qset.filter(created__lte=age)
        #     for slog in qset:
        #         slog.logout()

    def notifyMobile(self, data=None, title=None, body=None):
        for device in self.devices.filter(cm_provider="fcm", state=1):
            if device.cm_token is None:
                continue
            if data is not None:
                device.sendData(data)
            else:
                device.sendNotification(title, body)

    def notify(self, template=None, context=None, subject=None, message=None, email_only=True, sms_msg=None, force=False, from_email=settings.DEFAULT_FROM_EMAIL):
        from telephony.models import SMS
        # do not allow if account is not active
        if not self.is_active and not force:
            return False
        via = self.getProperty("notify_via", "all")
        phone = self.getProperty("phone")
        email = self.email
        valid_email = email is not None and "@" in email and "invalid" not in email
        allow_sms = not email_only and phone and (force or via in ["all", "sms"])
        allow_email = valid_email and (force or via in ["all", "email"])
        if not allow_email and not allow_sms:
            return False

        if allow_email:
            ctx = {
                'to': self.email,
                'to_token': crypto.hashit(self.email),
                'from': from_email,
                'timezone': "America/Los_Angeles"
            }

            if context:
                subject = context.get("subject", subject)
                message = context.get("message", message)
                ctx.update(context)

            from_email = ctx.get("from", None)
            rest_mail.send(
                [self.email],
                subject,
                message,
                from_email=from_email,
                template=template,
                context=ctx,
                do_async=True
            )
            self.log("notified", subject, method=self.email)

        if allow_sms:
            if not sms_msg and subject:
                sms_msg = subject
            if not sms_msg and message:
                sms_msg = message
            SMS.send(phone, sms_msg)
            self.log("notified", subject, method=phone)
        return True

    def checkIsOwner(self, member):
        return self.id == member.id

    def canSee(self, user):
        if user.id == self.id or self.hasPerm(["manage_members", "view_members"]):
            return True
        return False

    def canEdit(self, user, allow_self_edit=True):
        if allow_self_edit and user.id == self.id:
            return True
        if user.is_staff:
            # only super user can edit staff users
            return self.is_superuser or self.hasPerm("manage_staff")
        if self.hasPerm(["manage_users", "manage_staff"]):
            return True
        # we need to find any groups these users have in common
        # then find if the request.member is a manager or hasPerm admin or manager
        common_groups = user.getGroups().filter(memberships__member=self)
        if common_groups.count():
            # we have some common groups lets see if self is a admin or manager
            qset = self.memberships.filter(group__in=common_groups).filter(Q(permissions__name__in=["admin", "manager", "manage_members"])|Q(role__icontains="manager"))
            return qset.count() > 0
        return False

    def canEditMe(self, user):
        return user.canEdit(self)

    def refreshSecurityToken(self):
        self.security_token = crypto.randomString(8)
        self.save()

    def generateAuthCode(self, length=6, expires=AUTH_CODE_EXPIRES_SECS):
        for i in range(0, 100):
            code = crypto.randomCode(length)
            if Member.objects.filter(auth_code=code).count() == 0:
                self.auth_code = code
                self.auth_code_expires = datetime.now() + timedelta(seconds=expires)
                self.save()
                return self.auth_code
        return None

    # time based one time passwords / GOOGLE Authenticator
    def totp_getSecret(self, reset=False):
        token = self.getProperty("totp_token", category="secrets", default=None)
        if token is None or reset:
            token = pyotp.random_base32()
            self.setProperty("totp_token", token, category="secrets")
        return token

    def totp_getURI(self):
        # this should only be used one time during setup
        token = self.totp_getSecret(reset=True)
        totp = pyotp.TOTP(token)
        return totp.provisioning_uri(name=self.username, issuer_name=settings.SITE_LABEL)

    def totp_verify(self, code, window=1):
        token = self.totp_getSecret()
        totp = pyotp.TOTP(token)
        return totp.verify(code, valid_window=window)

    def reportIncident(self, category, description, level=10, request=None, background=False, **kwargs):
        incident_event = incident.event
        if background is False:
            incident_event = incident.event_now
        if request is None:
            request = self.getActiveRequest()
        incident_event(
            category, description, level,
            request=request,
            username=self.username,
            component="account.Member",
            component_id=self.pk,
            **kwargs)

    def block(self, reason, request=None):
        if not request:
            request = self.getActiveRequest()
        PersistentLog.log("account blocked, {}".format(reason), 5, request, "account.Member", self.pk, "blocked")
        RemoteEvents.hset("users:blocked:username", self.username, time.time())
        self.reportIncident(
            "account", f"account blocked: {reason}", level=2,
            request=request,
            details=reason)

    def unblock(self, request=None):
        if not request:
            request = self.getActiveRequest()
        if request and request.user.is_authenticated:
            who = request.user.username
        else:
            who = "time"
        PersistentLog.log("account unblocked by {}".format(who), 5, request, "account.Member", self.pk, "unblocked")
        RemoteEvents.hdel("users:blocked:username", self.username)
        RemoteEvents.hdel("users:failed:username", self.username)

    def getActiveConnections(self):
        c = RemoteEvents.hget("member:online:connections", self.id)
        if c:
            return int(c)
        return 0

    # BEGIN REST CALLBACKS
    def on_permission_change(self, key, value, old_value, category):
        # called when a metadata.permissions field changes.. see django settings USER_METADATA_PROPERTIES 
        # we want to log both the person changing permissions
        # and those being changed
        request = RestModel.getActiveRequest()
        perm = key
        reason = ""

        if "reason" in request.DATA:
            reason = request.DATA.get("reason").strip()

        if request.member:
            if value in [None, 0, '0']:
                self.log("remove_perm", "{} removed perm {}; {}".format(request.user.username, perm, reason), method="permissions", level=10)
                request.member.log("removed_perm", "removed perm {} for {}; {}".format(perm, self.username, reason), method="permissions", level=10)
            else:
                self.log("add_perm", "{} added perm {}; {}".format(request.user.username, perm, reason), method="permissions", level=10)
                request.member.log("gave_perm", "gave perm {} for {}; ".format(perm, self.username, reason), method="permissions", level=10)
        else:
            if value in [None, 0, '0']:
                self.log("remove_perm", "system removed perm {}".format(perm), method="permissions", level=10)
            else:
                self.log("add_perm", "system added perm {}".format(perm), method="permissions", level=10)

    def set_is_staff(self, value):
        request = self.getActiveRequest()
        if not request.member.hasPermission("manage_staff"):
            raise PermissionDeniedException("Permission Denied: attempting to set staff user")
        self.is_staff = int(value)

    def set_is_superuser(self, value):
        request = self.getActiveRequest()
        if not request.member.is_superuser:
            raise PermissionDeniedException("Permission Denied: attempting to super user")
        self.is_superuser = int(value)     

    def set_disable(self, value):
        if value is not None and value in [1, '1', True, 'true']:
            self.set_action("enable")
        self.set_action("disable")

    def set_action(self, value):
        action = value
        request = self.getActiveRequest()
        if action in ["unlock", "unblock"]:
            if not request.member.is_staff and not self.canEditMe(request.member):
                raise PermissionDeniedException("Permission Denied: attempting to unlock user")
            self.unblock(request)
        elif action == "disable":
            if self.is_superuser and not request.user.is_superuser:
                raise PermissionDeniedException("Permission Denied: attempting to disable super user")
            if self.is_staff and not request.member.is_staff:
                raise PermissionDeniedException("Permission Denied: attempting to disable staff user")
            if not request.member.is_staff and not self.canEditMe(request.member):
                raise PermissionDeniedException("Permission Denied: attempting to dsiable user")
            self.disable(request.user)
        elif action == "enable":
            if not request.member.is_staff and not self.canEditMe(request.member):
                raise PermissionDeniedException("Permission Denied: attempting to enable user")
            self.enable(request.member)
        elif action == "touch_password":
            if not request.member.is_staff and not self.canEditMe(request.member):
                raise PermissionDeniedException("Permission Denied: attempting to touch password")
            self.password_changed = datetime.now()
            self.save()
        elif action == "update_password_next":
            # force the user to update password on next login
            if not request.member.is_staff and not self.canEditMe(request.member):
                raise PermissionDeniedException("Permission Denied: attempting to touch password")
            self.password_changed = datetime.now() - timedelta(days=settings.PASSWORD_EXPIRES_DAYS - 1)
            self.save()
        elif action == "auth_token":
            if not request.member.is_superuser and not self.canEditMe(request.member):
                raise PermissionDeniedException("Permission Denied: attempting to generate auth_token")
            token = self.default_auth_token
            if token is None:
                token = AuthToken(member=self, role="default")
            token.generateToken()

    def set_full_name(self, value):
        self.set_name(value)

    def set_name(self, value):
        if not value:
            return
        names = value.split(' ')
        self.first_name = names[0].title()
        if len(names) > 1:
            self.last_name = " ".join(names[1:]).title()
        self.display_name = value.title()

    def set_username(self, value, generate=True):
        # we force our usernames to be the sames as the email
        value = value.strip()
        value = value.lower()
        value = value.replace(' ', '.')
        if '@' in value:
            uname = self.username
            self.username = None
            self.set_email(value)
            if self.username is None:
                self.username = uname
        elif self.username != value:
            orig_value = value
            if generate:
                for i in range(0, 20):
                    if Member.verifyUsername(value, self.id):
                        self.username = value
                        return True
                    value = "{}{}".format(orig_value, i)
            else:
                if Member.verifyUsername(value, self.id):
                    self.username = value
                    return True
            raise RestValidationError("username '{}'' already exists!".format(value))

    def set_newpassword(self, value):
        request = self.getActiveRequest()
        if not request:
            raise RestValidationError("requires request to continue")
        old_password = request.DATA.get("oldpassword", None)
        if not old_password:
            request.member.log("password_error", "requires oldpassword to change password", method="password_change", level=10)
            raise RestValidationError("requires oldpassword to continue")
        # verify we have the old password correct
        if not self.checkPassword(old_password):
            # invalid password
            request.member.log("password_error", "incorrect oldpassword to change password", method="password_change", level=10)
            raise RestValidationError("old password is not correct")
        self.set_password(value)

    def set_password(self, value):
        """
        this is tricky because we need to call set_password on the User model
        """
        request = self.getActiveRequest()
        if not request.member.canEdit(self):
            request.member.log("permission_denied", "attempting to set password for user: {}".format(self.username), method="password_change", level=3)
            raise PermissionDeniedException("Permission Denied: attempting to change password")
        if request.member.id != self.id:
            self.log("modified_by", "password changed by: {}".format(request.member.username), method="password_change", level=31)
            request.member.log("member_edit", "{} password changed".format(self.username), method="password_change", level=31)
            self.setPassword(value, skip_history=True)
        else:
            self.setPassword(value)

    def set_email(self, value):
        # we force our usernames to be the sames as the email
        # basic validation
        if value is not None:
            value = value.lower().strip()

        if self.email == value:
            return

        if "@" not in value or "." not in value:
            raise RestValidationError("Invalid Email")

        # verify there is not another account with this email
        qs = Member.objects.filter(email=value)
        if self.id:
            qs.exclude(pk=self.id)
        if qs.count():
            raise RestValidationError("user with email {} already exists!".format(value))

        if self.email:
            msg = f"email changed from {self.email} to {value}"
            self.reportIncident(
                "account", msg, level=8,
                details=msg)
            self.log("email_changed", msg, method="email_change", level=7)
        self.email = value

        if self.username:
            if "@" in self.username:
                if Member.verifyUsername(value, self.id):
                    self.username = value
            return True

        if len(value) > 250:
            value = value.split("@")[0]
        if Member.verifyUsername(value, self.id):
            self.username = value
        else:
            raise RestValidationError("email to username '{}' already exists!".format(value))

    def set_invite_body(self, value):
        # this is called typically for a system level user
        if value:
            subject = "Invite Link"
            request = self.getActiveRequest()
            if request is not None:
                subject = request.DATA.get("invite_subject", subject)
            self.sendEmail(subject, value)

    def on_rest_can_get(self, request):
        if request.member.pk == self.pk:
            return True
        return super().on_rest_can_get(request)

    def on_rest_can_save(self, request):
        if request.member.pk == self.pk:
            return True
        return super().on_rest_can_save(request)

    # def on_rest_pre_save(self, request, **kwargs):
    #     if not self.pk:
    #         q = Q(username=self.username.strip())
    #         if self.email:
    #             q = q | Q(email=self.email.strip())
    #         if Member.objects.filter(q).count() > 0:
    #             raise PermissionDeniedException(f"user {self.username} already exists")

    def on_rest_created(self, request):
        self.uuid = Member.generateUUID(self.pk)
        if not self.display_name:
            if self.full_name:
                self.display_name = self.full_name
            else:
                self.display_name = self.username
                if "@" in self.display_name:
                    self.display_name = self.display_name[:self.display_name.find("@")]
        self.save()
        created_by = "system"
        if request is not None and request.member:
            created_by = request.member.username
        self.auditLog(f"created by {created_by}", "created", level=21)

    # BEGIN STATIC METHODS
    @staticmethod
    def getByUser(user):
        member = Member.objects.filter(pk=user.pk).first()
        if member is None:
            member = Member(user_ptr=user)
            for f in user._meta.local_fields: setattr(member, f.name, getattr(user, f.name))
            member.save()
        return member

    @staticmethod
    def RecordInvalidLogin(request, username=None):
        # this records events when a username doesn't even exist
        if not username:
            username = request.DATA.get("username", None)
        if username:
            RemoteEvents.hset("users:failed:username", username, 0)
            incident.event_now("account", f"unknown user: {username}", level=3, request=request, error="invalid_login")

    @staticmethod
    def verifyUsername(username, exclude=None):
        qs = Member.objects.filter(username=username)
        if exclude:
            qs.exclude(pk=exclude)
        return qs.count() == 0

    @staticmethod
    def GetMember(username):
        if "@" in username:
            m = Member.objects.filter(email=username.lower()).last()
            if m:
                return m
        return Member.objects.filter(username=username.lower()).last()

    @staticmethod
    def GetMemberByPhone(phone_number):
        phone_number = rh.normalizePhone(phone_number)
        return Member.objects.filter(phone_number=phone_number.lower()).last()

    @staticmethod
    def GetWithPermission(perm, email_list=False):
        if type(perm) is list:
            queries = [Q(properties__category="permissions", properties__key=p, properties__value="1") for p in perm]
            query = queries.pop()
            for item in queries:
                query |= item
            qset = Member.objects.filter(is_active=True).filter(query).distinct()
        else:
            qset = Member.objects.filter(is_active=True).filter(properties__category="permissions", properties__key=perm, properties__value="1")

        if email_list:
            return list(qset.exclude(email__icontains="invalid").values_list('email', flat=True))
        return qset

    @staticmethod
    def GetWithNotification(perm, email_list=False):
        if isinstance(perm, list):
            perms = []
            for p in perm:
                if p.startswith("notify."):
                    p = p.split(".")[1]
                perms.append(p)
            queries = [Q(properties__category="notify", properties__key=p, properties__int_value=1) for p in perms]
            query = queries.pop()
            for item in queries:
                query |= item
            qset = Member.objects.filter(is_active=True).filter(query).distinct()
        else:
            if perm.startswith("notify."):
                perm = perm.split(".")[1]
            qset = Member.objects.filter(is_active=True).filter(properties__category="notify", properties__key=perm, properties__int_value=1)

        if email_list:
            return list(qset.exclude(email__icontains="invalid").values_list('email', flat=True))
        return qset

    @staticmethod
    def FilterOnline(is_online, qset):
        ids = list(RemoteEvents.smembers("member:online"))
        if is_online:
            qset = qset.filter(pk__in=ids)
        else:
            qset = qset.exclude(pk__in=ids)
        return qset

    @staticmethod
    def FilterBlocked(is_blocked, qset):
        ids = list(RemoteEvents.hgetall("users:blocked:username"))
        if is_blocked:
            qset = qset.filter(username__in=ids)
        else:
            qset = qset.exclude(username__in=ids)
        return qset

    @classmethod
    def on_rest_handle_filter(cls, qset, filters, request=None):
        if "is_blocked" in filters:
            if request.default_rest_filters:
                j = request.default_rest_filters.pop("is_blocked", None)
            qset = cls.FilterBlocked(filters.pop("is_blocked") in [1, "1", True], qset)
        if "is_online" in filters:
            if request.default_rest_filters:
                j = request.default_rest_filters.pop("is_online", None)
            qset = cls.FilterOnline(filters.pop("is_online") in [1, "1", True], qset)
        if "perm" in filters:
            perm = filters.pop("perm", None)
            if perm is not None:
                qset = qset.filter(
                    properties__category="permissions",
                    properties__key=perm,
                    properties__int_value=1)
        return qset

    @staticmethod
    def notifyWithPermission(perm, subject, message=None, template=None, context=None, email_only=False, sms_msg=None, force=False):
        NotificationRecord = RestModel.getModel("account", "NotificationRecord")
        NotificationRecord.notify(Member.GetWithPermission(perm), subject, message, template, context, email_only, sms_msg, force)

    @staticmethod
    def notifyWith(setting, subject, message=None, template=None, context=None, email_only=False, sms_msg=None, force=False):
        NotificationRecord = RestModel.getModel("account", "NotificationRecord")
        NotificationRecord.notify(Member.GetWithNotification(setting), subject, message, template, context, email_only, sms_msg, force)

    @classmethod
    def authWS4RedisConnection(cls, auth_data):
        """
        This method is used by the async/websocket service to authenticate.
        If the model can authenticate the connection it should return dict 
        with kind and pk of the model that is authenticaed
        """
        from rest import UberDict
        member = cls.getMemberFromSessionStore(auth_data.token)
        if member is not None:
            return UberDict(kind="member", pk=member.pk, member=member)
        return None

    @classmethod
    def getMemberFromSessionStore(cls, key):
        # assume this is a session key
        # this is deprecated, only using for backwards
        from importlib import import_module
        engine = import_module(settings.SESSION_ENGINE)
        session = engine.SessionStore(key)
        if "member_id" in session:
            return cls.objects.filter(pk=session["member_id"]).last()
        return None

    @classmethod
    def getMemberFromSession(cls, key):
        session = Session.objects.filter(session_key=key).last()
        if session is None:
            rh.log_print("no session for key", key)
            return None
        session_data = session.get_decoded()
        rh.log_print(session_data)
        uid = session_data.get('_auth_user_id', None)
        if uid is None:
            return None
        return cls.objects.filter(pk=uid).last()


class MemberMetaData(MetaDataBase):
    parent = models.ForeignKey(Member, related_name="properties", on_delete=models.CASCADE)


# class MemberPermission(models.Model):
#     member = models.ForeignKey(Member, related_name="permissions", on_delete=models.CASCADE)
#     name = models.CharField(max_length=255, db_index=True)


class PasswordHistory(models.Model):
    created = models.DateTimeField(auto_now_add=True, editable=False)
    owner = models.ForeignKey(Member, related_name="password_history", on_delete=models.CASCADE)
    password = models.CharField(max_length=255)

    @staticmethod
    def HashPassword(password):
        return hashlib.sha512((settings.SECRET_KEY + password).encode('utf-8')).hexdigest()


class AuthToken(models.Model, RestModel):
    class RestMeta:
        NO_SHOW_FIELDS = ["token"]
        CAN_DELETE = True
        SEARCH_FIELDS = ["member__username", "member__email", "member__first_name", "member__last_name"]
        GRAPHS = {
            "default": {
                "extra": ["secure_token"],
                "graphs": {
                    "member": "simple"
                }
            },
            "list": {
                "extra": ["secure_token"],
                "graphs": {
                    "member": "simple"
                }
            }
        }
    created = models.DateTimeField(auto_now_add=True, editable=False)
    token = models.TextField(db_index=True, unique=True)
    member = models.ForeignKey(Member, related_name="auth_tokens", on_delete=models.CASCADE)
    role = models.CharField(max_length=128, null=True, default=None, blank=True)
    signature = models.CharField(max_length=128, null=True, default=None, blank=True, db_index=True)

    def generateToken(self, commit=True):
        self.token = str(uuid.uuid1())
        self.updateSignature()
        if commit:
            self.save()

    def updateSignature(self):
        self.signature = crypto.hashSHA256(self.token)

    def on_rest_pre_save(self, request, **kwargs):
        if not self.token:
            self.token = uuid.uuid4().hex
            self.updateSignature()
        if not self.member_id:
            self.member = request.member
        if not request.member.canEdit(self.member):
            raise PermissionDeniedException("user is not allowed to create auth token for user", 469)

    @property
    def secure_token(self):
        request = self.getActiveRequest()
        if request:
            if self.member == request.member or request.member.is_superuser:
                return self.token
            if len(self.token) > 6:
                return "{}{}".format("*" * (len(self.token)-4), self.token[-4:])
        return "{}{}".format("*" * (len(self.token)-4), self.token[-4:])

    def __str__(self):
        return "{o.member}: {o.secure_token}".format(o=self)





# Generated by Django 4.1.4 on 2023-07-13 15:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0012_settings_settingsmetadata'),
    ]

    operations = [
        migrations.AddField(
            model_name='memberdevice',
            name='ip',
            field=models.CharField(default=None, max_length=64, null=True),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-29 20:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0005_member_auth_code_expires'),
    ]

    operations = [
        migrations.AddField(
            model_name='member',
            name='security_token',
            field=models.CharField(blank=True, db_index=True, default=None, max_length=64, null=True),
        ),
    ]
